#!/usr/bin/env python

"""Debian/Ubuntu package release."""

import os
import shutil
import sys

from nuitka.options.CommandLineOptionsTools import makeOptionsParser
from nuitka.tools.release.DebianPackage import (
    checkChangeLog,
    cleanupTarfileForDebian,
    runPy2dsc,
    shallNotIncludeInlineCopy,
)
from nuitka.tools.release.Release import (
    checkBranchName,
    cleanupSourceDistributionState,
    getBranchCategory,
)
from nuitka.Tracing import my_print, tools_logger
from nuitka.utils.Execution import check_call, withEnvironmentVarOverridden
from nuitka.utils.FileOperations import (
    copyTree,
    getFileList,
    listDir,
    putTextFileContents,
    resolveShellPatternToFilenames,
    withDirectoryChange,
)
from nuitka.Version import getNuitkaVersion


def parseArgs():
    parser = makeOptionsParser(usage=None, epilog=None)

    parser.add_option(
        "--no-pbuilder-update",
        action="store_false",
        dest="update_pbuilder",
        default=True,
        help="""\
Update the pbuilder chroot before building. Default %default.""",
    )

    parser.add_option(
        "--check",
        action="store_true",
        dest="check",
        default=False,
        help="""\
Check only, if the package builds, do not upload. Default %default.""",
    )

    parser.add_option(
        "--src",
        action="store",
        dest="src",
        default="/src",
        help="""\
Where the package source resides. Default %default.""",
    )

    options, positional_args = parser.parse_args()

    assert len(positional_args) == 1, positional_args
    codename = positional_args[0]

    return options, codename


def fixupPermissionsInplace(
    dest_path,
    file_perm,
    dir_perm,
    ignore_filenames=(),
):
    for filename in getFileList(dest_path, ignore_filenames=ignore_filenames):
        os.chmod(filename, file_perm)
        os.chmod(os.path.dirname(filename), dir_perm)


def fixupPermissionsCopy(source_path, dest_path, ignore_filenames=()):
    copyTree(source_path=source_path, dest_path=dest_path)

    fixupPermissionsInplace(
        dest_path=dest_path,
        file_perm=int("600", 8),
        dir_perm=int("700", 8),
        ignore_filenames=ignore_filenames,
    )


def main():
    # Complex stuff, pylint: disable=too-many-statements

    # Make sure error messages are in English.
    os.environ["LC_ALL"] = "C"

    options, codename = parseArgs()

    nuitka_version = getNuitkaVersion()

    branch_name = checkBranchName()

    category = getBranchCategory(branch_name)

    if category == "stable":
        if nuitka_version.count(".") == 1:
            assert checkChangeLog("New upstream release.")
        else:
            assert checkChangeLog("New upstream hotfix release.")
    else:
        assert checkChangeLog("New upstream pre-release.")

    # Clean the stage for the debian package. The name "deb_dist" is what "py2dsc"
    # uses for its output later on.
    shutil.rmtree("dist", ignore_errors=True)
    shutil.rmtree("build", ignore_errors=True)

    tools_logger.info(
        "Copying src to temporary location before modifying it.", style="blue"
    )
    copyTree(options.src, "/var/tmp/src")
    os.chdir("/var/tmp/src")

    cleanupSourceDistributionState()

    fixupPermissionsInplace(
        "debian", int("644", 8), int("755", 8), ignore_filenames=("rules",)
    )
    fixupPermissionsInplace("nuitka", int("644", 8), int("755", 8))

    # Build the source distribution as per our setup, however that's not going
    # to be good enough for Debian packages. We may want to avoid inline copies
    # being added to sdist for the Debian that do not need it.
    # spell-checker: ignore gztar
    with withEnvironmentVarOverridden(
        "NUITKA_NO_INLINE_COPY",
        "1" if shallNotIncludeInlineCopy(codename) else "0",
    ):
        tools_logger.info("Creating dist archive.", style="blue")
        check_call([sys.executable, "setup.py", "sdist", "--formats=gztar"])

        with withDirectoryChange("dist"):
            # Provide a re-packed tar.gz for the Debian package as input.

            # Create it as a "+ds" file, removing:
            # - the benchmarks (too many sources, not useful to end users, potential license
            #   issues)
            # - the inline copy of scons (not wanted for Debian)

            for _fullname, filename in listDir("."):
                if filename.endswith(".tar.gz"):
                    new_name = filename[:-7] + "+ds.tar.gz"

                    cleanupTarfileForDebian(
                        codename=codename, filename=filename, new_name=new_name
                    )

                    tools_logger.info("Running py2dsc.", style="blue")

                    entry = runPy2dsc(filename, new_name)

                    break
            else:
                assert False

        # Copy gnupg folder, to control the permissions.
        fixupPermissionsCopy(source_path="/root/.gnupg", dest_path="/root/.gnupg2")
        os.environ["GNUPGHOME"] = "/root/.gnupg2"

        # Copy ssh folder to control permissions.
        fixupPermissionsCopy(source_path="/root/.ssh2", dest_path="/root/.ssh")

        with withDirectoryChange(os.path.join("dist", "deb_dist", entry)):
            for filename in getFileList("debian", ignore_filenames=("rules",)):
                os.chmod(filename, int("644", 8))

            # Build the debian package, but disable the running of tests, will be done later
            # in the pbuilder test steps, pass inline copy setting on, and make sure that
            # the signing works, gpg-agent otherwise refuses to start.
            check_call(
                [
                    "debuild",
                    "--set-envvar=DEB_BUILD_OPTIONS=nocheck",
                    "--set-envvar=NUITKA_NO_INLINE_COPY=%s"
                    % os.environ["NUITKA_NO_INLINE_COPY"],
                    "--set-envvar=GNUPGHOME=%s" % os.environ["GNUPGHOME"],
                ]
            )

    # Check with pylint in pedantic mode and don't proceed if there were any
    # warnings given. Nuitka is lintian clean and shall remain that way. For
    # hotfix releases, i.e. "stable" builds, we skip this test though.
    if category == "stable":
        my_print("Skipped lintian checks for stable releases.", style="blue")
    else:
        # TODO: Do the lintian checks inside the pbuilder, we do not build
        # the package locally anymore, or with a builder that focusses on
        # that with e.g. the develop branch as a target on a regular basis.
        my_print("Skipped lintian checks for now.", style="blue")
        # assert os.system("lintian --pedantic --allow-root dist/deb_dist/*.changes") == 0

    # Move the created debian package files out.
    assert os.system("cp dist/deb_dist/*.deb dist/") == 0

    # Build inside the pbuilder chroot, and output to dedicated directory.
    shutil.rmtree("package", ignore_errors=True)
    os.makedirs("package")

    # Now update the pbuilder, spell-checker: ignore basetgz

    basetgz_filename = "/pbuilder/%s.tgz" % codename

    if options.update_pbuilder:
        command = """\
sudo /usr/sbin/pbuilder --update --basetgz %s""" % (basetgz_filename)

        check_call(
            ["sudo", "/usr/sbin/pbuilder", "--update", "--basetgz", basetgz_filename]
        )

    (dsc_filename,) = resolveShellPatternToFilenames("dist/deb_dist/*.dsc")

    # Execute the package build in the pbuilder with tests.
    # spell-checker: ignore hookdir,debemail,buildresult
    command = (
        "sudo",
        "/usr/sbin/pbuilder",
        "--build",
        "--basetgz",
        basetgz_filename,
        "--hookdir",
        "debian/pbuilder-hookdir",
        "--debemail",
        "Kay Hayen <kay.hayen@gmail.com>",
        "--buildresult",
        "package",
        dsc_filename,
    )

    check_call(command, shell=False)

    # Cleanup the build directory, not needed anymore.
    shutil.rmtree("build", ignore_errors=True)

    # Now build the repository.
    my_print("Building repository ...", style="blue")

    os.chdir("package")

    os.makedirs("repo")
    os.chdir("repo")

    os.makedirs("conf")

    # spell-checker: ignore armel armhf
    putTextFileContents(
        "conf/distributions",
        contents="""\
Origin: Nuitka
Label: Nuitka
Codename: %(codename)s
Architectures: i386 amd64 armel armhf powerpc
Components: main
Description: Apt repository for project Nuitka %(codename)s
SignWith: D96ADCA1377F1CEB6B5103F11BFC33752912B99C
""" % {"codename": codename},
    )

    # spell-checker: ignore reprepro includedeb
    command = ["reprepro", "includedeb", codename]
    command.extend(resolveShellPatternToFilenames("../*.deb"))

    check_call(command, shell=False)

    if options.check:
        tools_logger.sysexit("Finished check.", exit_code=0)

    my_print("Uploading ...", style="blue")

    # Create repo folder unless already done. This is needed for the first
    # build only. It also prevents us from talking to foreign servers, by
    # letting ssh check all it wants.
    assert (
        os.system(
            "ssh root@ssh.nuitka.net mkdir -p /var/www/deb/%s/%s/"
            % (category, codename)
        )
        == 0
    )

    # Update the repository on the web site.
    assert (
        os.system(
            "rsync -avz --delete dists pool --chown www-data root@ssh.nuitka.net:/var/www/deb/%s/%s/"
            % (category, codename)
        )
        == 0
    )

    my_print("Finished.", style="blue")



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Utilities for private pip space to download packages into."""

import os
import re
import subprocess
import sys

from nuitka.__past__ import total_ordering
from nuitka.containers.OrderedDicts import OrderedDict
from nuitka.PythonVersions import (
    getSitePackageCandidateNames,
    getSystemPrefixPath,
    python_version_str,
)

from .Download import getDownloadCacheDir, shouldDownload
from .Execution import (
    NuitkaCalledProcessError,
    check_output,
    getExecutablePath,
    getNullOutput,
    withEnvironmentPathAdded,
)
from .FileOperations import (
    areSamePaths,
    deleteFile,
    getFileContentByLine,
    getNormalizedPathJoin,
    putBinaryFileContents,
    removeDirectory,
    withTemporaryFilename,
)
from .Hashing import HashCRC32
from .Importing import withTemporarySysPathExtension
from .Utils import getArchitecture, getOS, isWin32Windows


def getSystemPrefixExecutable():
    """Return the sys.executable of the system prefix, i.e. breaking out of virtualenv.

    See getSystemPrefixPath() for details.

    Returns:
        str - path to system executable
    """
    # Yes, we want and need it here, pylint: disable=protected-access
    if hasattr(sys, "_base_executable"):
        return sys._base_executable

    sys_prefix = getSystemPrefixPath()
    if sys_prefix == sys.prefix:
        return sys.executable

    if areSamePaths(sys_prefix, sys.prefix):
        return sys.executable

    # The "sys.executable" is a good guess for the name, but we might have to
    # look for "python" effectively.
    candidate_names = (
        os.path.basename(sys.executable),
        "python.exe" if os.name == "nt" else "python",
        "python%s" % python_version_str,
    )

    for candidate_name in candidate_names:
        if os.name == "nt":
            candidate = os.path.join(sys_prefix, candidate_name)
        else:
            candidate = os.path.join(sys_prefix, "bin", candidate_name)

        if os.path.exists(candidate):
            return candidate

    # Fallback to sys.executable, which is not correct if we are in a virtualenv
    # but the best we can do if we didn't find the system prefix one.
    return sys.executable


def _getCandidateBinPaths(logger, site_packages):
    """Get the candidate binary paths for the private pip space."""
    download_folder = getPrivatePipBaseFolder()

    candidate_bin_paths = [
        os.path.join(download_folder, "bin"),
        os.path.join(download_folder, "usr", "bin"),
        os.path.join(download_folder, "usr", "local", "bin"),
    ]

    if isWin32Windows():
        candidate_bin_paths.insert(0, os.path.join(download_folder, "Scripts"))

    if site_packages is None:
        site_packages = getPrivatePipSitePackagesDir(logger=logger)

    if site_packages:
        if isWin32Windows():
            candidate_bin_paths.insert(
                0,
                os.path.join(
                    os.path.dirname(os.path.dirname(site_packages)), "Scripts"
                ),
            )
        else:
            candidate_bin_paths.insert(
                0,
                os.path.join(
                    os.path.dirname(os.path.dirname(os.path.dirname(site_packages))),
                    "bin",
                ),
            )

    return candidate_bin_paths


def _normalizePrivatePipPackageName(package_name):
    """Normalize a package name to the dist-info naming convention."""
    return re.sub(r"[-_.]+", "_", package_name)


def _clearPrivatePipBinaryCheckCache(binary_path):
    """Invalidate cached checks for a private pip binary path."""
    _check_required_version_cache.pop((binary_path, "--version"), None)
    _check_mdformat_usability_cache.pop(binary_path, None)


def _removePrivatePipInstalledPath(logger, path):
    """Remove a private pip installation path if it exists."""
    if not os.path.lexists(path):
        return False

    if os.path.isdir(path) and not os.path.islink(path):
        removeDirectory(
            path=path,
            logger=logger,
            ignore_errors=False,
            extra_recommendation="Please remove the private pip cache folder manually and retry",
        )
    else:
        deleteFile(path, must_exist=False)

    return True


def _cleanupPrivatePipModuleState(logger, site_packages_folder, module_name):
    """Remove stale module files for a package refresh."""
    removed = False

    if site_packages_folder is not None and os.path.isdir(site_packages_folder):
        module_path = getNormalizedPathJoin(
            site_packages_folder, module_name.replace(".", os.path.sep)
        )

        for candidate in (
            module_path,
            module_path + ".py",
            module_path + ".pyc",
            module_path + ".pyo",
        ):
            if _removePrivatePipInstalledPath(logger=logger, path=candidate):
                removed = True

        pycache_dir = getNormalizedPathJoin(os.path.dirname(module_path), "__pycache__")
        if os.path.isdir(pycache_dir):
            module_basename = os.path.basename(module_path)

            for filename in os.listdir(pycache_dir):
                if filename.startswith(module_basename + ".") and filename.endswith(
                    ".pyc"
                ):
                    if _removePrivatePipInstalledPath(
                        logger=logger,
                        path=os.path.join(pycache_dir, filename),
                    ):
                        removed = True

    return removed


def _cleanupPrivatePipMetadataState(logger, site_packages_folder, package_name):
    """Remove stale package metadata for a package refresh."""
    removed = False

    if site_packages_folder is not None and os.path.isdir(site_packages_folder):
        dist_info_prefix = _normalizePrivatePipPackageName(package_name) + "-"

        for filename in os.listdir(site_packages_folder):
            if filename.startswith(dist_info_prefix) and filename.endswith(
                (".dist-info", ".egg-info")
            ):
                if _removePrivatePipInstalledPath(
                    logger=logger,
                    path=os.path.join(site_packages_folder, filename),
                ):
                    removed = True

    return removed


def _getPrivatePipBinaryCleanupPaths(candidate_bin_path, binary_name):
    """Get the binary wrapper paths to replace during a refresh."""
    if isWin32Windows():
        return (
            os.path.join(candidate_bin_path, binary_name + ".exe"),
            os.path.join(candidate_bin_path, binary_name + ".exe.manifest"),
            os.path.join(candidate_bin_path, binary_name + ".cmd"),
            os.path.join(candidate_bin_path, binary_name + "-script.py"),
            os.path.join(candidate_bin_path, binary_name + "-script.pyw"),
        )
    else:
        return (os.path.join(candidate_bin_path, binary_name),)


def _cleanupPrivatePipBinaryState(logger, site_packages_folder, binary_names):
    """Remove stale binary wrappers for a package refresh."""
    removed = False
    seen_bin_paths = set()

    for candidate_bin_path in _getCandidateBinPaths(
        logger=logger, site_packages=site_packages_folder
    ):
        if candidate_bin_path in seen_bin_paths:
            continue

        seen_bin_paths.add(candidate_bin_path)

        for binary_name in binary_names:
            binary_paths = _getPrivatePipBinaryCleanupPaths(
                candidate_bin_path=candidate_bin_path,
                binary_name=binary_name,
            )

            _clearPrivatePipBinaryCheckCache(binary_paths[0])

            for binary_path in binary_paths:
                if _removePrivatePipInstalledPath(logger=logger, path=binary_path):
                    removed = True

    return removed


def _cleanupPrivatePipPackageState(
    logger,
    site_packages_folder,
    package_name,
    module_name,
    binary_names,
):
    """Remove stale package metadata, modules, and scripts before a refresh."""

    removed = _cleanupPrivatePipModuleState(
        logger=logger,
        site_packages_folder=site_packages_folder,
        module_name=module_name,
    )
    removed = (
        _cleanupPrivatePipMetadataState(
            logger=logger,
            site_packages_folder=site_packages_folder,
            package_name=package_name,
        )
        or removed
    )
    removed = (
        _cleanupPrivatePipBinaryState(
            logger=logger,
            site_packages_folder=site_packages_folder,
            binary_names=binary_names,
        )
        or removed
    )

    return removed


def _checkPrivatePipBinaryPath(
    logger,
    binary_name,
    binary_path,
    version,
    dependencies,
    check_binary,
    report_rejection,
):
    """Check if a binary path is version compatible and usable."""

    with withPrivatePipSitePackagesPathAdded(logger=logger):
        if version is None:
            ok = True
            message = None
        else:
            ok, message = _checkRequiredVersion(
                logger=logger,
                tool=binary_name,
                tool_call=[binary_path],
                dependencies=dependencies,
            )

        if ok and check_binary is not None:
            ok, message = check_binary(
                logger=logger,
                binary_path=binary_path,
            )

    if report_rejection and not ok and logger is not None:
        logger.info(
            "Rejecting '%s' binary '%s' due to: %s"
            % (binary_name, binary_path, message)
        )

    return ok, message


def _getPrivatePipBinaryPath(
    logger,
    binary_name,
    package_name,
    module_name,
    dependencies,
    check_binary,
    assume_yes_for_downloads,
    reject_message,
):
    """Get the path of a binary from the private pip space or globally.

    This first checks the system PATH for the binary. If found, it checks if its
    version and its dependencies' versions match the requirements. If they match,
    it returns the global binary.

    Otherwise, it downloads the package and its dependencies into the private pip
    space, verifies the installed binary there, and returns its path.
    """
    # Complex code, pylint: disable=too-many-locals

    version = getRequiredVersion(logger, package_name)
    report_rejection = reject_message is not None

    candidate_bin_paths = _getCandidateBinPaths(logger=logger, site_packages=None)

    # Construct an extra_dir for search
    extra_dir = os.pathsep.join(candidate_bin_paths)

    binary_path = getExecutablePath(binary_name, extra_dir=extra_dir)
    force_package_update = False

    if binary_path is not None:
        ok, _message = _checkPrivatePipBinaryPath(
            logger=logger,
            binary_name=binary_name,
            binary_path=binary_path,
            version=version,
            dependencies=dependencies,
            check_binary=check_binary,
            report_rejection=report_rejection,
        )

        if ok:
            return binary_path, None, assume_yes_for_downloads

        force_package_update = True

    if reject_message is None:
        return None, None, assume_yes_for_downloads

    if force_package_update:
        site_packages_folder = getPrivatePipSitePackagesDir(logger=logger)

        cleaned_private_pip = _cleanupPrivatePipPackageState(
            logger=logger,
            site_packages_folder=site_packages_folder,
            package_name=package_name,
            module_name=module_name,
            binary_names=(binary_name,),
        )

        for dep_package_name, dep_module_name, _dep_reject_message in dependencies:
            cleaned_private_pip = (
                _cleanupPrivatePipPackageState(
                    logger=logger,
                    site_packages_folder=site_packages_folder,
                    package_name=dep_package_name,
                    module_name=dep_module_name,
                    binary_names=(),
                )
                or cleaned_private_pip
            )

        if cleaned_private_pip and logger is not None:
            logger.info(
                "Removed stale private pip package state for '%s' before refresh."
                % binary_name
            )

    # Download, avoiding to use the result, which is the site-packages
    # folder
    site_packages_folder, assume_yes_for_downloads = tryDownloadPackageName(
        logger=logger,
        package_name=package_name,
        module_name=module_name,
        package_version=version,
        force_update=force_package_update,
        assume_yes_for_downloads=assume_yes_for_downloads,
        reject_message=reject_message,
    )

    if site_packages_folder:
        for dep_package_name, dep_module_name, dep_reject_message in dependencies:
            _, assume_yes_for_downloads = tryDownloadPackageName(
                logger=logger,
                package_name=dep_package_name,
                module_name=dep_module_name,
                package_version=getRequiredVersion(logger, dep_package_name),
                force_update=force_package_update,
                assume_yes_for_downloads=assume_yes_for_downloads,
                reject_message=dep_reject_message,
            )

        # Check standard locations in the downloaded environment.
        for candidate in _getCandidateBinPaths(
            logger=logger, site_packages=site_packages_folder
        ):
            possible = os.path.join(candidate, binary_name)
            if isWin32Windows():
                possible += ".exe"

            if os.path.exists(possible):
                ok, _message = _checkPrivatePipBinaryPath(
                    logger=logger,
                    binary_name=binary_name,
                    binary_path=possible,
                    version=version,
                    dependencies=dependencies,
                    check_binary=check_binary,
                    report_rejection=report_rejection,
                )

                if not ok:
                    continue

                return possible, site_packages_folder, assume_yes_for_downloads

    return None, site_packages_folder, assume_yes_for_downloads


def getPrivatePipBaseFolder():
    """Get the inline copy folder for a given name."""
    return os.path.join(
        getDownloadCacheDir(),
        "pip",
        "private-%s" % _getPrivatePipHash(),
    )


_private_pip_site_packages_dir = None


def getPrivatePipSitePackagesDir(logger):
    """Get the site-packages directory of the private pip space."""
    # We use the global statement to cache the result across calls.
    # pylint: disable=global-statement
    global _private_pip_site_packages_dir

    if _private_pip_site_packages_dir is None:
        download_folder = getPrivatePipBaseFolder()

        for root, dirnames, _filenames in os.walk(download_folder):
            found_candidate = None

            for candidate in getSitePackageCandidateNames():
                if candidate in dirnames:
                    # Unclear which one to use.
                    if found_candidate is not None:
                        return logger.sysexit(
                            "Scan for pip folder found multiple candidates: %s and %s."
                            % (found_candidate, candidate)
                        )

                    found_candidate = candidate

            if found_candidate:
                _private_pip_site_packages_dir = os.path.join(root, found_candidate)
                break

    return _private_pip_site_packages_dir


def withPrivatePipSitePackagesPathAdded(logger):
    """Context manager to add private pip site-packages to PYTHONPATH."""
    return withEnvironmentPathAdded(
        "PYTHONPATH", getPrivatePipSitePackagesDir(logger=logger), prefix=True
    )


def _isPackageInstalled(site_packages_folder, package_name, package_version):
    """Check if a package is installed in the given site-packages folder."""
    if package_version is None:
        return True

    # Basic check for dist-info specific to version
    dist_info_name = "%s-%s" % (
        package_name.replace("-", "_"),
        package_version,
    )

    dist_info_path = getNormalizedPathJoin(
        site_packages_folder,
        "%s.dist-info" % dist_info_name,
    )
    egg_info_path = getNormalizedPathJoin(
        site_packages_folder,
        "%s.egg-info" % dist_info_name,
    )

    if not os.path.isdir(dist_info_path) and not os.path.isdir(egg_info_path):
        return False
    else:
        # Metadata found, but check for conflicts with other versions,
        # e.g. when an upgrade was done, but not fully clean.
        for filename in os.listdir(site_packages_folder):
            if (
                filename.startswith(package_name + "-")
                and (filename.endswith(".dist-info") or filename.endswith(".egg-info"))
                and filename != os.path.basename(dist_info_path)
                and filename != os.path.basename(egg_info_path)
            ):
                return False

    return True


def tryDownloadPackageName(
    logger,
    package_name,
    module_name,
    package_version,
    force_update,
    assume_yes_for_downloads,
    reject_message,
):
    """Try to download a package from the private pip space."""
    download_folder = getPrivatePipBaseFolder()

    site_packages_folder = getPrivatePipSitePackagesDir(logger=logger)

    if site_packages_folder is not None and not force_update:
        candidate = os.path.join(site_packages_folder, module_name)

        if os.path.exists(candidate):
            # If version is specified, check if it looks installed.
            if _isPackageInstalled(
                site_packages_folder=site_packages_folder,
                package_name=package_name,
                package_version=package_version,
            ):
                return (
                    site_packages_folder,
                    assume_yes_for_downloads,
                )

    if not _checkPackageConstraint(logger, package_name):
        return None, assume_yes_for_downloads

    if shouldDownload(
        message="Nuitka depends on '%s' to compile code, it is recommended."
        % package_name,
        reject_message=reject_message,
        assume_yes_for_downloads=assume_yes_for_downloads,
        download_ok=True,
    ):
        assume_yes_for_downloads = True

        if package_version is not None:
            package_spec = "%s==%s" % (package_name, package_version)
        else:
            package_spec = package_name

        command = (
            getSystemPrefixExecutable(),
            "-m",
            "pip",
            "install",
            "--no-warn-script-location",
            "--disable-pip-version-check",
            "--ignore-installed",
            "--upgrade",
            "--root",
            download_folder,
            "--prefix",
            ".",
            package_spec,
        )

        interactive = sys.stdout.isatty()

        with getNullOutput() as null_output:
            exit_code = subprocess.call(
                command,
                shell=False,
                stdout=None if interactive else null_output,
                stderr=None if interactive else null_output,
            )

        if exit_code != 0:
            return None, assume_yes_for_downloads

        if site_packages_folder is None:
            site_packages_folder = getPrivatePipSitePackagesDir(logger=logger)

        if site_packages_folder is None or not _isPackageInstalled(
            site_packages_folder=site_packages_folder,
            package_name=package_name,
            package_version=package_version,
        ):
            return logger.sysexit(
                "Error, failed to download package %r into private pip space."
                % package_name
            )

    if site_packages_folder is not None:
        return (
            site_packages_folder,
            assume_yes_for_downloads,
        )

    return None, assume_yes_for_downloads


def _evaluateConstraint(logger, constraint, package_name):
    """Evaluate a package constraint."""
    try:
        # We trust the file content, pylint: disable=eval-used
        return eval(
            constraint,
            {
                "python_version": _PythonVersion(python_version_str),
                "os": os,
                "sys": sys,
            },
        )
    except Exception as e:  # pylint: disable=broad-except
        return logger.sysexit(
            "Error, checking constraint %r for package %r gave error: %r"
            % (constraint, package_name, e)
        )


def _checkPackageConstraint(logger, package_name):
    """Check if a package constraint is met."""
    # Make sure we have the constraints loaded.
    _parseRequirements(logger)

    constraints = _private_pip_constraints.get(package_name)

    if not constraints:
        return True

    for constraint in constraints:
        if _evaluateConstraint(logger, constraint, package_name):
            return True

    return False


def getPrivatePackage(
    logger,
    package_name,
    module_name,
    package_version,
    submodule_names,
    assume_yes_for_downloads,
    reject_message,
):
    """Get a package from the private pip space or globally.

    Args:
        package_name: name of the package to download
        module_name: name of the module to import
        package_version: version of the package to download
        submodule_names: tuple of submodule names to import before returning
        assume_yes_for_downloads: if tools should be downloaded automatically

    Returns:
        module: the imported module or None if not found
    """
    try:
        module = __import__(module_name, fromlist=["*"])

        if submodule_names:
            for submodule_name in submodule_names:
                __import__(module_name + "." + submodule_name)

        return module
    except ImportError:
        site_packages_folder, _assume_yes_for_downloads = tryDownloadPackageName(
            logger=logger,
            package_name=package_name,
            module_name=module_name.split(".")[0],
            package_version=package_version,
            force_update=False,
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message=reject_message,
        )

        if site_packages_folder:
            with withTemporarySysPathExtension([site_packages_folder]):
                module = __import__(module_name, fromlist=["*"])

                if submodule_names:
                    for submodule_name in submodule_names:
                        __import__(module_name + "." + submodule_name)

            return module

        return None


def getZigBinaryPath(logger, assume_yes_for_downloads, reject_message):
    """Get the path of the Zig binary from the private pip space."""
    # spell-checker: ignore ziglang

    site_packages_folder, _assume_yes_for_downloads = tryDownloadPackageName(
        logger=logger,
        package_name="ziglang",
        module_name="ziglang",
        package_version=None,
        force_update=False,
        assume_yes_for_downloads=assume_yes_for_downloads,
        reject_message=reject_message,
    )

    # The "ziglang" package puts a "python-zig.exe" in the scripts folder, and
    # not found due to mismatch, but the actual "zig.exe" is in the site-packages
    # folder.
    if site_packages_folder:
        zig_exe_path = getNormalizedPathJoin(site_packages_folder, "ziglang", "zig")

        if isWin32Windows():
            zig_exe_path += ".exe"

        if os.path.exists(zig_exe_path):
            return zig_exe_path

    return None


def getClangFormatBinaryPath(logger, assume_yes_for_downloads, reject_message):
    """Get the path of the clang-format binary from the private pip space."""
    binary_path, _site_packages_folder, _assume_yes_for_downloads = (
        _getPrivatePipBinaryPath(
            logger=logger,
            binary_name="clang-format",
            package_name="clang-format",
            module_name="clang_format",
            dependencies=(),
            check_binary=None,
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message=reject_message,
        )
    )
    return binary_path


def getBlackBinaryPath(logger, assume_yes_for_downloads):
    """Get the path of the black binary from the private pip space."""
    binary_path, _site_packages_folder, _assume_yes_for_downloads = (
        _getPrivatePipBinaryPath(
            logger=logger,
            binary_name="black",
            package_name="black",
            module_name="black",
            dependencies=(),
            check_binary=None,
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message="Python formatting needs to use 'black'.",
        )
    )
    return binary_path


def getIsortBinaryPath(logger, assume_yes_for_downloads):
    """Get the path of the isort binary from the private pip space."""
    binary_path, _site_packages_folder, _assume_yes_for_downloads = (
        _getPrivatePipBinaryPath(
            logger=logger,
            binary_name="isort",
            package_name="isort",
            module_name="isort",
            dependencies=(),
            check_binary=None,
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message="Python formatting needs to use 'isort'.",
        )
    )
    return binary_path


def _checkMdformatUsability(logger, binary_path):
    """Check if mdformat can format a minimal markdown file."""
    # Not logging these calls, pylint: disable=unused-argument

    if binary_path in _check_mdformat_usability_cache:
        return _check_mdformat_usability_cache[binary_path]

    result = True, None

    with withTemporaryFilename(suffix=".md") as temp_filename:
        try:
            putBinaryFileContents(temp_filename, b"# smoke\n")
            check_output([binary_path, "--check", temp_filename])
        except NuitkaCalledProcessError as e:
            stderr = e.stderr

            if str is not bytes and stderr is not None:
                stderr = stderr.decode("utf8", "ignore")

            result = (
                False,
                "failed to format a test file: %s" % (stderr or e),
            )
        finally:
            deleteFile(temp_filename, must_exist=False)

    _check_mdformat_usability_cache[binary_path] = result

    return result


def getMdformatBinaryPath(logger, assume_yes_for_downloads):
    """Get the path of the mdformat binary from the private pip space."""

    dependencies = (
        (
            "mdformat-gfm",
            "mdformat_gfm",
            "Markdown formatting needs to use 'mdformat-gfm'.",
        ),
        (
            "mdformat-frontmatter",
            "mdformat_frontmatter",
            "Markdown formatting needs to use 'mdformat-frontmatter'.",
        ),
        (
            "mdformat-footnote",
            "mdformat_footnote",
            "Markdown formatting needs to use 'mdformat-footnote'.",
        ),
    )

    mdformat_path, _site_packages_folder, _assume_yes_for_downloads = (
        _getPrivatePipBinaryPath(
            logger=logger,
            binary_name="mdformat",
            package_name="mdformat",
            module_name="mdformat",
            dependencies=dependencies,
            check_binary=_checkMdformatUsability,
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message="Markdown formatting needs to use 'mdformat'.",
        )
    )

    return mdformat_path


def getRstfmtBinaryPath(logger, assume_yes_for_downloads):
    """Get the path of the rstfmt binary from the private pip space."""
    binary_path, _site_packages_folder, _assume_yes_for_downloads = (
        _getPrivatePipBinaryPath(
            logger=logger,
            binary_name="rstfmt",
            package_name="rstfmt",
            module_name="rstfmt",
            dependencies=(),
            check_binary=None,
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message="ReStructuredText formatting needs to use 'rstfmt'.",
        )
    )
    return binary_path


@total_ordering
class _PythonVersion(object):
    def __init__(self, version_str):
        self.version = tuple(int(x) for x in version_str.split("."))

    @staticmethod
    def _makeTuple(other):
        if str is not bytes and isinstance(other, (str, bytes)):
            return tuple(int(x) for x in other.split("."))
        return other

    def __eq__(self, other):
        return self.version == self._makeTuple(other)

    def __lt__(self, other):
        return self.version < self._makeTuple(other)


_private_pip_requirements = None
_private_pip_constraints = None


def _parseRequirements(logger):
    """Parse the requirements-private.txt file."""
    # We use the global statement to cache the result across calls.
    # pylint: disable=global-statement
    global _private_pip_requirements, _private_pip_constraints

    if _private_pip_requirements is None:
        _private_pip_requirements = OrderedDict()
        _private_pip_constraints = OrderedDict()

        for line in getFileContentByLine(
            os.path.join(os.path.dirname(__file__), "requirements-private.txt")
        ):
            line = line.strip()

            if not line or line.startswith("#"):
                continue

            parts = re.split(r"\s*;\s*", line, maxsplit=1)

            if len(parts) > 1:
                tool_part = parts[0]
                constraint_part = parts[1]
            else:
                tool_part = line
                constraint_part = None

            if "==" in tool_part:
                tool_name = tool_part.split("==")[0].strip()
            else:
                for other_comparison in (">=", "<=", ">", "<", "~="):
                    if other_comparison in tool_part:
                        msg = (
                            """\
Error, invalid requirement format in 'requirements-private.txt' line %r, only '==' is allowed."""
                            % line
                        )

                        return logger.sysexit(msg)

                tool_name = tool_part.strip()

            if tool_name not in _private_pip_requirements:
                _private_pip_requirements[tool_name] = []
                _private_pip_constraints[tool_name] = []

            _private_pip_requirements[tool_name].append(line)
            if constraint_part is not None:
                _private_pip_constraints[tool_name].append(constraint_part)

    return _private_pip_requirements


def _getPrivatePipHash():
    """Calculate the hash of the private pip space."""
    private_pip_hash = HashCRC32()
    private_pip_hash.updateFromValues(
        python_version_str,
        getOS(),
        getArchitecture(),
    )
    private_pip_hash.updateFromValues(_parseRequirements(logger=None))

    return private_pip_hash.asHexDigest()


def getRequiredVersion(logger, tool):
    """Get the required version of a tool from requirements-private.txt."""
    requirements = _parseRequirements(logger)

    if tool not in requirements:
        return logger.sysexit(
            "Error, cannot find %r in 'requirements-private.txt' file." % tool
        )

    for line in requirements[tool]:
        parts = re.split(r"\s*;\s*", line, maxsplit=1)

        if len(parts) > 1:
            constraint = parts[1]
            if not _evaluateConstraint(logger, constraint, tool):
                continue

        tool_part = parts[0]

        # TODO: This is not quite correct for all packages, but we only have equality
        # checks right now.
        if "==" in tool_part:
            return tool_part.split("==")[1].strip()
        else:
            return None

    return None


_check_required_version_cache = {}
_check_mdformat_usability_cache = {}


def _checkRequiredVersion(logger, tool, tool_call, dependencies):
    """Check if the version of a tool matches the requirement."""

    # Complex code, pylint: disable=too-many-branches,too-many-locals
    required_version = getRequiredVersion(logger=logger, tool=tool)

    tool_call = tuple(list(tool_call) + ["--version"])

    if tool_call in _check_required_version_cache:
        return _check_required_version_cache[tool_call]

    try:
        version_output = check_output(tool_call)
    except NuitkaCalledProcessError as e:
        result = False, "failed to execute: %s" % e.stderr
        _check_required_version_cache[tool_call] = result
        return result

    if str is not bytes:
        version_output = version_output.decode("utf8")

    if dependencies:
        for dep_package_name, _dep_module_name, _dep_reject_message in dependencies:
            plugin_ver = getRequiredVersion(logger, dep_package_name)
            if plugin_ver:
                search_name_1 = dep_package_name
                search_name_2 = dep_package_name.replace("-", "_")

                if not any(
                    "%s %s" % (name, plugin_ver) in version_output
                    for name in (search_name_1, search_name_2)
                ):
                    result = (
                        False,
                        "Missing or mismatched dependency '%s %s' in '%s'."
                        % (dep_package_name, plugin_ver, tool),
                    )
                    _check_required_version_cache[tool_call] = result
                    return result

    actual_version = None
    for line in version_output.splitlines():
        line = line.strip()

        if line.startswith(
            ("black, ", "python -m black,", "__main__.py, ", "mdformat ")
        ):
            if "(" in line:
                line = line[: line.rfind("(")].strip()

            actual_version = line.split()[-1]
            break
        if line.startswith("VERSION "):
            actual_version = line.split()[-1]
            break
        # spell-checker: ignore rstfmt
        if line.startswith("rstfmt "):
            actual_version = line.split()[-1]
            break
        if line.startswith("clang-format version "):
            actual_version = line.split()[2]
            break

    if not actual_version:
        result = False, "Error, couldn't determine version output of '%s' ('%s')" % (
            tool,
            " ".join(tool_call),
        )
        _check_required_version_cache[tool_call] = result
        return result

    message = "Version of '%s' via '%s' is required to be %r and not %r." % (
        tool,
        " ".join(tool_call),
        required_version,
        actual_version,
    )

    result = required_version == actual_version, message
    _check_required_version_cache[tool_call] = result
    return result


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Generate header files that provide compiler specifics."""

import os

from nuitka.build.AdaptPythonHeaderFiles import getOffsetsJsonRequiredKeys
from nuitka.build.SconsInterface import (
    cleanSconsDirectory,
    getCommonSconsOptions,
    runScons,
)
from nuitka.PythonVersions import isPythonWithGil, python_version_str
from nuitka.tools.quality.auto_format.AutoFormat import (
    withFileOpenedAndAutoFormatted,
)
from nuitka.Tracing import offsets_logger
from nuitka.utils.Execution import check_output
from nuitka.utils.FileOperations import (
    getNormalizedPath,
    makeContainingPath,
    makePath,
    putTextFileContents,
    withTemporaryFilename,
)
from nuitka.utils.Jinja2 import getTemplateC
from nuitka.utils.PrivatePipSpace import getZigBinaryPath
from nuitka.utils.Utils import getArchitecture, getOS


def getPythonInternalsOffsetDir():
    """Return the source distribution bundled JSON storage path resolving strictly against this module."""

    return getNormalizedPath(
        os.path.join(
            os.path.dirname(__file__),
            "..",
            "..",
            "..",
            "build",
            "python_internal_offset",
        )
    )


def generateHeader():
    scons_options, env_values = getCommonSconsOptions()

    if getOS() == "Windows":
        scons_options["zig_exe_path"] = getZigBinaryPath(
            logger=offsets_logger,
            assume_yes_for_downloads=True,
            reject_message="Nuitka needs zig for header generation.",
        )
        if scons_options["zig_exe_path"] is None:
            return offsets_logger.sysexit("Nuitka needs zig for header generation.")

    scons_options["source_dir"] = "generate_header.build"
    cleanSconsDirectory(scons_options["source_dir"])
    makePath(scons_options["source_dir"])

    keys = getOffsetsJsonRequiredKeys(python_version_str)

    template = getTemplateC(
        package_name="nuitka.tools.general.generate_header",
        template_name="GenerateHeadersMain.c.j2",
    )

    c_code = template.render(keys=keys)

    c_filepath = os.path.join(
        scons_options["source_dir"],
        "static_src",
        "GenerateHeadersMain.c",
    )
    makeContainingPath(c_filepath)
    putTextFileContents(c_filepath, c_code)

    python_version_id = "%s-%s-%s-%s" % (
        python_version_str,
        getOS(),
        getArchitecture(),
        "gil" if isPythonWithGil() else "no-gil",
    )

    with withTemporaryFilename(prefix=python_version_id, suffix=".exe") as result_exe:
        scons_options["result_exe"] = result_exe

        success = runScons(
            scons_options=scons_options,
            env_values=env_values,
            scons_filename="Offsets.scons",
        )

        if not success:
            return offsets_logger.sysexit(
                "Error, failed to compile offsets generation program."
            )

        header_output = check_output([result_exe])

        if str is not bytes:
            header_output = header_output.decode("utf8")

        lines = [line.strip() for line in header_output.splitlines() if line.strip()]

        json_filename = os.path.join(
            getPythonInternalsOffsetDir(),
            "offsets_%s.json" % python_version_id,
        )

        makeContainingPath(json_filename)

        with withFileOpenedAndAutoFormatted(
            json_filename, effective_filename=json_filename
        ) as output_f:
            output_f.write("".join(lines))

        offsets_logger.info("Gathered offsets into %s" % json_filename)
        offsets_logger.info("OK.")


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


import itertools


def calledRepeatedly(cond):
    if cond:
        # At function entry, before first assignment "NUITKA_ILONG_UNASSIGNED" which
        # means no C or object value is value.
        i = 0
        # When assigning from constants, we do NUITKA_ILONG_BOTH_VALID. Depending
        # on the type, we might not

        # Comparing to a CLONG, is easy. Actually for such small values, a DIGIT
        # type will be used, which is like a CLONG but smaller and known to work
        # well with "PyLongObject *" internals.
        while i < 20000:  # RICH_COMPARE_CBOOL_NILONG_CLONG(i, 9)
            # BINARY_OPERATION_ADD_NILONG_NILONG_CLONG(i, 1)
            i = i + 1
            # NUITKA_ILONG_CLONG_VALID

        # The flag NUITKA_INT_OBJECT_VALID is not set, need to create the object
        # pointer and return it.
        return i


for x in itertools.repeat(None, 50):
    # construct_begin
    calledRepeatedly(True)
    # construct_alternative
    calledRepeatedly(False)
    # construct_end

print("OK.")

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""User plugin exercising post-load code on a multi-phase extension module."""

from nuitka.plugins.PluginBase import NuitkaPluginBase


class NuitkaPluginForTesting(NuitkaPluginBase):
    plugin_name = "post-load-multiphase"

    @staticmethod
    def createPostModuleLoadCode(module):
        # spell-checker: ignore testmultiphase
        if module.getFullName() == "_testmultiphase":
            return (
                """\
from __future__ import absolute_import
from _testmultiphase import Example

assert Example.__name__ == "Example"
""",
                """\
Verify post-load code for multi-phase extension modules runs after execution.""",
            )


#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Interface to pefile on Windows."""

import os

from nuitka.containers.OrderedSets import OrderedSet
from nuitka.utils.FileOperations import listDllFilesFromDirectory
from nuitka.utils.SharedLibraries import getPEFileUsedDllNames

from .DllDependenciesCommon import (
    addMissingMsvcRedistDLL,
    attemptToFindNotFoundDLL,
    isMsvcRedistDllName,
    reportMissingMsvcRedistDLLs,
    shallIgnoreMissingDLL,
)

_scan_dir_dll_cache = {}


def _getScanDirectoryDLLs(scan_dirs):
    cache_key = tuple(scan_dirs)

    if cache_key not in _scan_dir_dll_cache:
        result = {}

        for scan_dir in scan_dirs:
            for dll_filename, dll_basename in listDllFilesFromDirectory(scan_dir):
                dll_basename = dll_basename.lower()

                if dll_basename not in result:
                    result[dll_basename] = os.path.normcase(
                        os.path.abspath(dll_filename)
                    )

        _scan_dir_dll_cache[cache_key] = result

    return _scan_dir_dll_cache[cache_key]


def detectDLLsWithPEFile(binary_filename, scan_dirs):
    """Detect DLLs used by a binary using recursive PE import analysis.

    Args:
        binary_filename: The binary to check.
        scan_dirs: Directories to search for DLLs.

    Returns:
        OrderedSet: Set of found DLL filenames.
    """
    result = OrderedSet()

    scan_dir_dlls = _getScanDirectoryDLLs(scan_dirs=scan_dirs)
    pending = [os.path.normcase(os.path.abspath(binary_filename))]
    scanned = set()

    while pending:
        current_binary = pending.pop()

        if current_binary in scanned:
            continue

        scanned.add(current_binary)

        pe_dll_names = getPEFileUsedDllNames(current_binary)

        if pe_dll_names is None:
            continue

        for dll_name in pe_dll_names:
            dll_name = dll_name.lower()

            dll_filename = scan_dir_dlls.get(dll_name)

            if dll_filename is None and isMsvcRedistDllName(dll_name):
                addMissingMsvcRedistDLL(dll_name)

            if (
                dll_filename is None
                and dll_name.startswith("python")
                and dll_name.endswith(".dll")
            ):
                dll_filename = attemptToFindNotFoundDLL(dll_name)

            if dll_filename is None:
                if shallIgnoreMissingDLL(dll_name):
                    continue

                continue

            dll_filename = os.path.normcase(os.path.abspath(dll_filename))
            dll_basename = os.path.basename(dll_filename).lower()

            if shallIgnoreMissingDLL(dll_basename):
                continue

            result.add(dll_filename)

            if dll_filename not in scanned:
                pending.append(dll_filename)

    reportMissingMsvcRedistDLLs()

    return result


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Test Python 3.14 function '__annotate__' behavior."""

from functools import singledispatch


def displayDict(d):
    result = "{"
    first = True
    for key, value in sorted(d.items()):
        if not first:
            result += ","

        result += "%s: %s" % (repr(key), repr(value))
        first = False
    result += "}"

    return result


def annotationCacheTest(_x: int):
    pass


if annotationCacheTest.__annotations__:
    pass

annotationCacheTest.__annotate__ = None
annotationCacheTest.__annotations__["cached"] = 9
print(
    "Annotation cache after clearing __annotate__:",
    displayDict(annotationCacheTest.__annotations__),
)


@singledispatch
def singledispatchTest(_value):
    return "default"


@singledispatchTest.register
def _(value: int):
    return "int"


print(
    "singledispatch register with annotation:",
    singledispatchTest(1),
    singledispatchTest("x"),
)


# Test class-level deferred annotations


class TestClassAnnotations:
    x: int = 1
    y: str

    def method(self, a: int) -> str:
        return str(a)


print(
    "Class annotations:",
    displayDict(TestClassAnnotations.__annotations__),
)
print("Class annotation attribute x:", TestClassAnnotations.x)
print(
    "Class annotation method result:",
    TestClassAnnotations().method(42),
)


class TestNoClassAnnotations:
    z = 42

    def method(self, a: int) -> str:
        return str(a)


print(
    "Class annotations (no class-level):",
    displayDict(TestNoClassAnnotations.__annotations__),
)
print("Class attribute z:", TestNoClassAnnotations.z)


class TestParent:
    parent_attr: str = "hello"


class TestChild(TestParent):
    child_attr: int = 10


print(
    "Child annotations:",
    displayDict(TestChild.__annotations__),
)
print("Child attr:", TestChild.child_attr)
print("Parent attr:", TestChild.parent_attr)

#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Cleanup of caches for Nuitka.

This is triggered by "--clean-cache=" usage, and can cleanup all kinds of
caches and is supposed to run before or instead of Nuitka compilation.
"""

import os

from nuitka.Tracing import cache_logger
from nuitka.utils.AppDirs import getCacheDir, removeCacheDir


def _cleanCacheDirectory(cache_name, cache_basename):
    from nuitka.options.Options import shallCleanCache

    if shallCleanCache(cache_name):
        cache_dir = getCacheDir(cache_basename, create=False)

        if os.path.exists(cache_dir):
            cache_logger.info(
                "Cleaning cache '%s' directory '%s'." % (cache_name, cache_dir)
            )
            removeCacheDir(
                cache_basename=cache_basename,
                logger=cache_logger,
                ignore_errors=False,
                extra_recommendation=None,
            )
            cache_logger.info("Done.")


def cleanCaches():
    _cleanCacheDirectory("ccache", "ccache")
    _cleanCacheDirectory("clcache", "clcache")
    _cleanCacheDirectory("zig", "zig")
    _cleanCacheDirectory("bytecode", "module-cache")
    _cleanCacheDirectory("dll-dependencies", "library_dependencies")


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


import os
import sys


def main():
    print("example_uv: main called")
    exit_code = 0

    # Check data file
    # We expect data.txt to be in the package directory
    import example_uv

    package_dir = os.path.dirname(example_uv.__file__)
    data_path = os.path.join(package_dir, "data.txt")

    if os.path.exists(data_path):
        print("Data file found.")
        with open(data_path, "rb") as f:
            print("Data content:", f.read().strip())
    else:
        print("Data file NOT found at:", data_path)
        exit_code = 1

    # Check example_uv.subpackage data files
    from example_uv import subpackage

    package_dir_2 = os.path.dirname(subpackage.__file__)
    included_path = os.path.join(package_dir_2, "data_included.txt")

    if not os.path.exists(included_path):
        print("Error: data_included.txt missing in example_uv.subpackage")
        exit_code = 1

    print("Data file inclusion verification passed.")

    sys.exit(exit_code)


if __name__ == "__main__":
    main()

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Release tools related to Debian packaging."""

import os
import sys

from nuitka.utils.Execution import check_call, getNullInput
from nuitka.utils.FileOperations import (
    copyFile,
    getFileContentByLine,
    getFileContents,
    listDir,
    openTextFile,
)


def _callDebchange(*args):
    # spell-checker: ignore debchange
    args = ["debchange"] + list(args)

    os.environ["EDITOR"] = ""

    with getNullInput() as null_input:
        check_call(args, stdin=null_input)


_discarded_last_entry = False


def _discardDebianChangelogLastEntry():
    # Kind of weak, but we do not keep track if we do a lot of things,
    # for every program run this is only done once to simplify code,
    # but it's not reusable therefore.
    # pylint: disable=global-statement
    global _discarded_last_entry

    if _discarded_last_entry:
        return
    changelog_lines = getFileContents("debian/changelog").splitlines()

    with openTextFile("debian/changelog", "w") as output:
        first = True
        for line in changelog_lines[1:]:
            if line.startswith("nuitka") and first:
                first = False

            if not first:
                output.write(line + "\n")

    _discarded_last_entry = True


def updateDebianChangelog(old_version, new_version, distribution):
    debian_version = new_version.replace("rc", "~rc") + "+ds-1"

    if old_version == new_version:
        _discardDebianChangelogLastEntry()

    os.environ["DEBEMAIL"] = "Kay Hayen <kay.hayen@gmail.com>"

    if "rc" in new_version:
        if "rc" in old_version:
            # Initial final release after pre-releases.
            _discardDebianChangelogLastEntry()

        message = "New upstream pre-release."
    else:
        if "rc" in old_version:
            # Initial final release after pre-releases.
            _discardDebianChangelogLastEntry()

            message = "New upstream release."
        else:
            # Initial final release after pre-releases.

            # Hotfix release after previous final or hotfix release.
            message = "New upstream hotfix release."

    _callDebchange("--newversion=%s" % debian_version, message)
    _callDebchange("-r", "--distribution", distribution, "")


def checkChangeLog(message):
    """Check debian changelog for given message to be present."""

    for line in getFileContentByLine("debian/changelog"):
        if line.startswith(" --"):
            return False

        if message in line:
            return True

    sys.exit("Error, didn't find in debian/changelog: '%s'" % message)


def shallNotIncludeInlineCopy(codename):
    # Keep inline copy for old variants that cannot otherwise work
    return codename not in ("stretch", "jessie", "xenial")


def cleanupTarfileForDebian(codename, filename, new_name):
    """Remove files that shouldn't be in Debian.

    The inline copies should definitely not be there. Also remove the
    PDF files.
    """

    copyFile(filename, new_name)
    check_call(["gunzip", new_name])

    if shallNotIncludeInlineCopy(codename):
        command = [
            "tar",
            "--wildcards",
            "--delete",
            "--file",
            new_name[:-3],
            "Nuitka*/build/inline_copy",
        ]

        check_call(command)

    check_call(["gzip", "-9", "-n", new_name[:-3]])


def runPy2dsc(filename, new_name):
    check_call(["py2dsc", new_name])

    # Fixup for py2dsc not taking our custom suffix into account, so we need
    # to rename it ourselves.
    before_deb_name = filename[:-7].lower().replace("-", "_")
    after_deb_name = before_deb_name.replace("rc", "~rc")

    os.rename(
        "deb_dist/%s.orig.tar.gz" % before_deb_name,
        "deb_dist/%s+ds.orig.tar.gz" % after_deb_name,
    )

    check_call(["rm -f deb_dist/*_source*"], shell=True)

    # Remove the now useless input, py2dsc has copied it, and we don't
    # publish it.
    os.unlink(new_name)

    # Assert that the unpacked directory is there and find it. Otherwise fail badly.

    entry = None
    for fullname, entry in listDir("deb_dist"):
        if (
            os.path.isdir(fullname)
            and entry.startswith("nuitka")
            and not entry.endswith(".orig")
        ):
            break

    if entry is None:
        assert False

    # Import the "debian" directory from above. It's not in the original tar and
    # overrides fully what py2dsc did.
    check_call(["rm -r deb_dist/%s/debian/*" % entry], shell=True)

    check_call(
        [
            "rsync",
            "-a",
            "--exclude",
            "pbuilder-hookdir",
            "../debian/",
            "deb_dist/%s/debian/" % entry,
        ]
    )

    check_call(["rm deb_dist/*.dsc deb_dist/*.debian.tar.xz"], shell=True)

    return entry


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Runner for generated tests of Nuitka.

These tests are created on the fly, and some use Nuitka internals to
decide what to test for or how, e.g. to check that a type indeed does
have a certain slot.

"""

import os
import sys

# Find nuitka package relative to us.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
    ),
)

# isort:start

import nuitka.specs.BuiltinStrOperationSpecs
from nuitka.tools.specialize.SpecializePython import (
    python2_dict_methods as dict_method_names,
)
from nuitka.tools.specialize.SpecializePython import (
    python2_str_methods as str_method_names,
)
from nuitka.tools.specialize.SpecializePython import (
    python3_bytes_methods as bytes_method_names,
)
from nuitka.tools.testing.Common import (
    compareWithCPython,
    createSearchMode,
    my_print,
    scanDirectoryForTestCases,
    setup,
)

# For templating.
operations = (
    ("Add", "+"),
    ("Sub", "-"),
    ("Pow", "**"),
    ("Mult", "*"),
    ("FloorDiv", "//"),
    ("Div", "/"),
    ("Mod", "%"),
    ("Pow", "**"),
    ("LShift", "<<"),
    ("RShift", ">>"),
    ("BitAnd", "&"),
    ("BitOr", "|"),
    ("BitXor", "^"),
    ("Divmod", "divmod"),
    ("Subscript", "["),
)


# For typical constant values to use in operation tests.
candidates = (
    ("NoneType", "None", "None"),
    ("bool", "True", "False"),
    ("int", "17", "-9"),
    ("float", "17.2", "-8"),
    ("complex", "2j", "-4j"),
    ("str", "'lala'", "'lol'"),
    ("bytearray", "bytearray(b'lulu')", "bytearray(b'lol')"),
    ("list", "[1,2]", "[3]"),
    ("tuple", "(1,2)", "(3,)"),
    ("set", "set([1,2])", "set([3])"),
    ("frozenset", "frozenset([1,2])", "frozenset([3])"),
    ("dict", "{1:2}", "{3:4}"),
)


# For making an operator usage, needed because divmod is function style.
def makeOperatorUsage(operator, left, right):
    if operator == "divmod":
        return "divmod(%s, %s)" % (left, right)
    elif operator == "[":
        return "%s[%s]" % (left, right)
    else:
        return "%s %s %s" % (left, operator, right)


def main():
    python_version = setup(suite="generated", needs_io_encoding=True)

    search_mode = createSearchMode()

    # Singleton, pylint: disable=global-statement
    global operations
    global candidates

    if python_version >= (3, 5):
        operations += (("MatMult", "@"),)

    if python_version < (3,):
        candidates += (("long", "17L", "-9L"),)
        candidates += (("unicode", "u'lala'", "u'lol'"),)
    else:
        candidates += (("bytes", "b'lala'", "b'lol'"),)

    method_arguments = {}

    for str_method_name in str_method_names:
        spec = getattr(
            nuitka.specs.BuiltinStrOperationSpecs, "str_%s_spec" % str_method_name, None
        )

        if spec is None:
            my_print(
                "Warning, str function '%s' has no spec." % str_method_name,
                style="yellow",
            )
            continue

        method_arguments[str_method_name] = spec.getArgumentNames()

    for bytes_method_name in bytes_method_names:
        spec = getattr(
            nuitka.specs.BuiltinBytesOperationSpecs,
            "bytes_%s_spec" % bytes_method_name,
            None,
        )

        if spec is None:
            my_print(
                "Warning, bytes function '%s' has no spec." % bytes_method_name,
                style="yellow",
            )
            continue

        method_arguments[bytes_method_name] = spec.getArgumentNames()

    for dict_method_name in dict_method_names:
        spec = getattr(
            nuitka.specs.BuiltinDictOperationSpecs,
            "dict_%s_spec" % dict_method_name,
            None,
        )

        if spec is None:
            my_print(
                "Warning, dict function '%s' has no spec." % dict_method_name,
                style="yellow",
            )
            continue

        method_arguments[dict_method_name] = spec.getArgumentNames()

    template_context = {
        "operations": operations,
        "dict_method_names": [
            dict_method_name
            for dict_method_name in dict_method_names
            if dict_method_name in method_arguments
        ],
        "str_method_names": [
            str_method_name
            for str_method_name in str_method_names
            if str_method_name in method_arguments
        ],
        "bytes_method_names": [
            bytes_method_name
            for bytes_method_name in bytes_method_names
            if bytes_method_name in method_arguments
        ],
        "method_arguments": method_arguments,
        "inplace_operations": tuple(
            operation
            for operation in operations
            if operation[0] not in ("Divmod", "Subscript")
        ),
        "candidates": candidates,
        "makeOperatorUsage": makeOperatorUsage,
        "len": len,
    }

    # Now run all the tests in this directory.
    for filename in scanDirectoryForTestCases(".", template_context=template_context):
        extra_flags = [
            # No error exits normally, unless we break tests, and that we would
            # like to know.
            "expect_success",
            # Keep no temporary files.
            "remove_output",
            # Do not follow imports.
            "--nofollow-imports",
            # Use the original __file__ value, at least one case warns about things
            # with filename included.
            "--file-reference-choice=original",
            # Cache the CPython results for reuse, they will normally not change.
            "cpython_cache",
        ]

        # This test should be run with the debug Python, and makes outputs to

        active = search_mode.consider(dirname=None, filename=filename)

        if active:
            compareWithCPython(
                dirname=None,
                filename=filename,
                extra_flags=extra_flags,
                search_mode=search_mode,
            )

    search_mode.finish()


if __name__ == "__main__":
    main()

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Utils module to provide helper for our common json operations."""

from __future__ import absolute_import

import json

from nuitka.__past__ import unicode

from .FileOperations import getFileContents, openTextFile, stripFileContentsBOM


def _convertJsonLoadedValues(value):
    # return driven, pylint: disable=too-many-return-statements

    if str is not bytes:
        return value

    if type(value) is unicode:
        try:
            return str(value)
        except UnicodeEncodeError:
            return value

    if type(value) is list:
        return [_convertJsonLoadedValues(element) for element in value]

    if type(value) is tuple:
        return tuple(_convertJsonLoadedValues(element) for element in value)

    if type(value) is dict:
        result = {}

        for key, element in value.items():
            result[_convertJsonLoadedValues(key)] = _convertJsonLoadedValues(element)

        return result

    return value


def loadJsonFromFilename(filename):
    try:
        contents = getFileContents(filename, mode="rb")
        contents, _bom = stripFileContentsBOM(contents)

        if type(contents) is bytes:
            contents = contents.decode("utf8")

        return _convertJsonLoadedValues(json.loads(contents))
    except ValueError:
        return None


def writeJsonToFile(file_handle, contents, indent=2):
    json.dump(contents, file_handle, indent=indent, sort_keys=True)
    file_handle.write("\n")


def writeJsonToFilename(filename, contents, indent=2):
    with openTextFile(filename, "w") as output:
        writeJsonToFile(output, contents, indent=indent)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""This test runner compiles all Python files as a module.

This is a test to achieve some coverage, it will only find assertions of
within Nuitka or warnings from the C compiler. Code will not be run
normally.

"""

import os
import sys

# Find nuitka package relative to us.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
    ),
)

# isort:start

import subprocess

from nuitka.tools.testing.Common import (
    checkCompilesNotWithCPython,
    compileLibraryTest,
    createSearchMode,
    getPythonArch,
    getPythonVendor,
    getTempDir,
    my_print,
    setup,
)
from nuitka.utils.Importing import getExtensionModuleSuffix

python_version = setup(suite="python_modules", needs_io_encoding=True)
python_vendor = getPythonVendor()
python_arch = getPythonArch()

search_mode = createSearchMode()

tmp_dir = getTempDir()

# spell-checker: ignore idnadata,tweedie

ignore_filenames_list = (
    "__phello__.foo.py",  # Triggers error for "." in module name
    "idnadata",  # Avoid too complex code for main program.
    "joined_strings.py",
    # Incredible amount of memory in C compiler for test code
    "test_spin.py",
    # Uses outside modules up the chain
    "cheshire_tomography.py",
)

late_syntax_errors = (
    # No syntax error with Python2 compileall, but run time only:
    "_identifier.py",
    "bench.py",
    "_tweedie_compound_poisson.py",
    "session.py",
)


def decide(_root, filename):
    return (
        filename.endswith(".py")
        and filename not in ignore_filenames_list
        and "(" not in filename
        and filename.count(".") == 1
    )


def action(stage_dir, _root, path):
    command = [
        sys.executable,
        os.path.join("..", "..", "bin", "nuitka"),
        "--mode=module",
        "--output-dir=%s" % stage_dir,
        "--remove-output",
        "--quiet",
        "--nofollow-imports",
        "--no-progressbar",
    ]

    command += os.environ.get("NUITKA_EXTRA_OPTIONS", "").split()

    suffix = getExtensionModuleSuffix(preferred=True)

    if os.path.basename(path) == "__init__.py":
        source_filename = os.path.dirname(path)
        target_filename = os.path.basename(source_filename) + suffix
    else:
        source_filename = path
        target_filename = os.path.basename(source_filename)[:-3] + suffix

    target_filename = target_filename.replace("(", "").replace(")", "")

    command.append(source_filename)

    try:
        subprocess.check_call(command)
    except subprocess.CalledProcessError:
        basename = os.path.basename(path)

        if basename in late_syntax_errors:
            my_print("Syntax error is known unreliable with file %s." % basename)
        else:
            my_print("Falling back to full comparison due to error exit.")

            checkCompilesNotWithCPython(
                dirname=None, filename=path, search_mode=search_mode
            )
    else:
        my_print("OK")

        os.unlink(os.path.join(stage_dir, target_filename))


compileLibraryTest(
    search_mode=search_mode,
    stage_dir=os.path.join(
        tmp_dir,
        "compile_library_%s-%s-%s"
        % (".".join(str(d) for d in python_version), python_arch, python_vendor),
    ),
    decide=decide,
    action=action,
)

search_mode.finish()

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Iteration related codes.

Next variants and unpacking with related checks.
"""

from nuitka.nodes.ConstantRefNodes import makeConstantRefNode
from nuitka.PythonVersions import python_version

from .CodeHelpers import (
    decideConversionCheckNeeded,
    generateChildExpressionsCode,
    generateExpressionCode,
    withObjectCodeTemporaryAssignment,
)
from .ComparisonCodes import getRichComparisonCode
from .ErrorCodes import (
    getErrorExitBoolCode,
    getErrorExitCode,
    getErrorExitReleaseCode,
    getFrameVariableTypeDescriptionCode,
    getReleaseCode,
)
from .LineNumberCodes import getErrorLineNumberUpdateCode
from .PythonAPICodes import generateCAPIObjectCode
from .templates.CodeTemplatesIterators import template_loop_break_next


def generateBuiltinNext1Code(to_name, expression, emit, context):
    (value_name,) = generateChildExpressionsCode(
        expression=expression, emit=emit, context=context
    )

    (
        exception_state_name,
        _exception_lineno,
    ) = context.variable_storage.getExceptionVariableDescriptions()

    with withObjectCodeTemporaryAssignment(
        to_name, "next_value", expression, emit, context
    ) as result_name:
        # TODO: Make use of "ITERATOR_NEXT_ITERATOR" in case of known type shape
        # iterator.
        emit(
            """\
%(to_name)s = ITERATOR_NEXT(%(iterator_name)s);
if (%(to_name)s == NULL) {
    FETCH_ERROR_OCCURRED_STATE(tstate, &%(exception_state_name)s);

    if (!HAS_EXCEPTION_STATE(&%(exception_state_name)s)) {
        SET_EXCEPTION_PRESERVATION_STATE_STOP_ITERATION_EMPTY(tstate, &%(exception_state_name)s);
    }
}
"""
            % {
                "to_name": result_name,
                "iterator_name": value_name,
                "exception_state_name": exception_state_name,
            }
        )

        getErrorExitCode(
            check_name=result_name,
            release_name=value_name,
            fetched_exception=True,
            emit=emit,
            context=context,
        )

        context.addCleanupTempName(result_name)


def getBuiltinLoopBreakNextCode(expression, to_name, value, emit, context):
    if expression.getTypeShape().isShapeIterator():
        emit("%s = %s;" % (to_name, "ITERATOR_NEXT_ITERATOR(%s)" % value))
    else:
        emit("%s = %s;" % (to_name, "ITERATOR_NEXT(%s)" % value))

    getReleaseCode(release_name=value, emit=emit, context=context)

    break_target = context.getLoopBreakTarget()
    if type(break_target) is tuple:
        break_indicator_code = "%s = true;" % break_target[1]
        break_target = break_target[0]
    else:
        break_indicator_code = ""

    (
        exception_state_name,
        _exception_lineno,
    ) = context.variable_storage.getExceptionVariableDescriptions()

    emit(
        template_loop_break_next
        % {
            "to_name": to_name,
            "break_indicator_code": break_indicator_code,
            "break_target": break_target,
            "release_temps": getErrorExitReleaseCode(context),
            "var_description_code": getFrameVariableTypeDescriptionCode(context),
            "line_number_code": getErrorLineNumberUpdateCode(context),
            "exception_target": context.getExceptionEscape(),
            "exception_state_name": exception_state_name,
        }
    )

    context.addCleanupTempName(to_name)


def generateSpecialUnpackCode(to_name, expression, emit, context):
    value_name = context.allocateTempName("unpack")

    generateExpressionCode(
        to_name=value_name,
        expression=expression.subnode_value,
        emit=emit,
        context=context,
    )

    with withObjectCodeTemporaryAssignment(
        to_name, "unpack_value", expression, emit, context
    ) as result_name:
        needs_check = expression.mayRaiseExceptionOperation()

        if not needs_check:
            emit("%s = UNPACK_NEXT_INFALLIBLE(%s);" % (result_name, value_name))
        elif python_version < 0x350:
            (
                exception_state_name,
                _exception_lineno,
            ) = context.variable_storage.getExceptionVariableDescriptions()

            emit(
                "%s = UNPACK_NEXT(tstate, &%s, %s, %s);"
                % (
                    result_name,
                    exception_state_name,
                    value_name,
                    expression.getCount() - 1,
                )
            )
        else:
            starred = expression.getStarred()
            expected = expression.getExpected()

            (
                exception_state_name,
                _exception_lineno,
            ) = context.variable_storage.getExceptionVariableDescriptions()

            emit(
                "%s = UNPACK_NEXT%s(tstate, &%s, %s, %s, %s);"
                % (
                    result_name,
                    "_STARRED" if starred else "",
                    exception_state_name,
                    value_name,
                    expression.getCount() - 1,
                    expected,
                )
            )

        getErrorExitCode(
            check_name=result_name,
            release_name=value_name,
            needs_check=needs_check,
            fetched_exception=True,
            emit=emit,
            context=context,
        )

    context.addCleanupTempName(to_name)


def generateUnpackCheckCode(statement, emit, context):
    iterator_name = context.allocateTempName("iterator_name")

    generateExpressionCode(
        to_name=iterator_name,
        expression=statement.subnode_iterator,
        emit=emit,
        context=context,
    )

    with context.withCurrentSourceCodeReference(statement.getSourceReference()):
        (
            exception_state_name,
            _exception_lineno,
        ) = context.variable_storage.getExceptionVariableDescriptions()

        res_name = context.getBoolResName()

        emit(
            "%s = UNPACK_ITERATOR_CHECK(tstate, &%s, %s, %d);"
            % (res_name, exception_state_name, iterator_name, statement.getCount())
        )

        getErrorExitBoolCode(
            condition="%s == false" % res_name,
            fetched_exception=True,
            emit=emit,
            context=context,
        )


def generateUnpackCheckFromIteratedCode(statement, emit, context):
    iteration_length_name = context.allocateTempName("iteration_length", unique=True)

    generateExpressionCode(
        to_name=iteration_length_name,
        expression=statement.subnode_iterated_length,
        emit=emit,
        context=context,
    )

    to_name = context.getBoolResName()

    getRichComparisonCode(
        to_name=to_name,
        comparator="Gt",
        left=statement.subnode_iterated_length,
        # Creating a temporary node on the fly, knowing it's not used for many
        # things. TODO: Once we have value shapes, we ought to use those.
        right=makeConstantRefNode(
            constant=statement.count,
            source_ref=statement.source_ref,
            user_provided=True,
        ),
        # We know that cannot fail.
        needs_check=False,
        source_ref=statement.source_ref,
        emit=emit,
        context=context,
    )

    # TODO: Why is this necessary, to_name doesn't allow storage.
    context.removeCleanupTempName(to_name)

    # TODO: This exception ought to have a creator function.
    emit("""
if (%(to_name)s) {
    PyErr_Format(PyExc_ValueError, "too many values to unpack");
}
""" % {"to_name": to_name})

    getErrorExitBoolCode(condition=str(to_name), emit=emit, context=context)


def generateBuiltinNext2Code(to_name, expression, emit, context):
    generateCAPIObjectCode(
        to_name=to_name,
        capi="BUILTIN_NEXT2",
        tstate=True,
        arg_desc=(
            ("next_arg", expression.subnode_iterator),
            ("next_default", expression.subnode_default),
        ),
        may_raise=expression.mayRaiseException(BaseException),
        conversion_check=decideConversionCheckNeeded(to_name, expression),
        source_ref=expression.getCompatibleSourceReference(),
        emit=emit,
        context=context,
    )


def generateBuiltinIter1Code(to_name, expression, emit, context):
    may_raise = expression.mayRaiseExceptionOperation()

    generateCAPIObjectCode(
        to_name=to_name,
        capi="MAKE_ITERATOR" if may_raise else "MAKE_ITERATOR_INFALLIBLE",
        tstate=may_raise,
        arg_desc=(("iter_arg", expression.subnode_value),),
        may_raise=may_raise,
        conversion_check=decideConversionCheckNeeded(to_name, expression),
        source_ref=expression.getCompatibleSourceReference(),
        emit=emit,
        context=context,
    )


def generateBuiltinIterForUnpackCode(to_name, expression, emit, context):
    may_raise = expression.mayRaiseExceptionOperation()

    generateCAPIObjectCode(
        to_name=to_name,
        capi="MAKE_UNPACK_ITERATOR" if may_raise else "MAKE_ITERATOR_INFALLIBLE",
        tstate=False,
        arg_desc=(("iter_arg", expression.subnode_value),),
        may_raise=may_raise,
        conversion_check=decideConversionCheckNeeded(to_name, expression),
        source_ref=expression.getCompatibleSourceReference(),
        emit=emit,
        context=context,
    )


def generateBuiltinIter2Code(to_name, expression, emit, context):
    generateCAPIObjectCode(
        to_name=to_name,
        capi="BUILTIN_ITER2",
        tstate=False,
        arg_desc=(
            ("iter_callable", expression.subnode_callable_arg),
            ("iter_sentinel", expression.subnode_sentinel),
        ),
        may_raise=expression.mayRaiseException(BaseException),
        conversion_check=decideConversionCheckNeeded(to_name, expression),
        source_ref=expression.getCompatibleSourceReference(),
        emit=emit,
        context=context,
    )


def generateBuiltinLenCode(to_name, expression, emit, context):
    generateCAPIObjectCode(
        to_name=to_name,
        capi="BUILTIN_LEN",
        tstate=True,
        arg_desc=(("len_arg", expression.subnode_value),),
        may_raise=expression.mayRaiseException(BaseException),
        conversion_check=decideConversionCheckNeeded(to_name, expression),
        source_ref=expression.getCompatibleSourceReference(),
        emit=emit,
        context=context,
    )


def generateBuiltinAnyCode(to_name, expression, emit, context):
    generateCAPIObjectCode(
        to_name=to_name,
        capi="BUILTIN_ANY",
        tstate=True,
        arg_desc=(("any_arg", expression.subnode_value),),
        may_raise=expression.mayRaiseException(BaseException),
        conversion_check=decideConversionCheckNeeded(to_name, expression),
        source_ref=expression.getCompatibleSourceReference(),
        emit=emit,
        context=context,
    )


def generateBuiltinAllCode(to_name, expression, emit, context):
    generateCAPIObjectCode(
        to_name=to_name,
        capi="BUILTIN_ALL",
        tstate=True,
        arg_desc=(("all_arg", expression.subnode_value),),
        may_raise=expression.mayRaiseException(BaseException),
        conversion_check=decideConversionCheckNeeded(to_name, expression),
        source_ref=expression.getCompatibleSourceReference(),
        emit=emit,
        context=context,
    )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



"""Debian/Ubuntu package release."""

import os
import shutil
import sys

from nuitka.tools.release.DebianPackage import (
    cleanupTarfileForDebian,
    runPy2dsc,
)
from nuitka.tools.release.Documentation import checkReleaseDocumentation
from nuitka.tools.release.Release import (
    checkBranchName,
    cleanupSourceDistributionState,
)
from nuitka.Tracing import my_print
from nuitka.utils.FileOperations import listDir


def main():
    branch_name = checkBranchName()
    assert branch_name == "main"

    shutil.rmtree("dist", ignore_errors=True)
    shutil.rmtree("build", ignore_errors=True)

    cleanupSourceDistributionState()

    checkReleaseDocumentation()

    assert os.system("%s setup.py sdist --formats=gztar" % sys.executable) == 0

    os.chdir("dist")

    # Clean the stage for the debian package. The name "deb_dist" is what "py2dsc"
    # uses for its output later on.
    shutil.rmtree("deb_dist", ignore_errors=True)

    # Provide a re-packed tar.gz for the Debian package as input.

    # Create it as a "+ds" file, removing:
    # - the benchmarks (too many sources, not useful to end users, potential license
    #   issues)
    # - the inline copy of scons (not wanted for Debian)

    # Then run "py2dsc" on it.

    for _fullname, filename in listDir("."):
        if filename.endswith(".tar.gz"):
            new_name = filename[:-7] + "+ds.tar.gz"

            cleanupTarfileForDebian(
                codename="sid",
                filename=filename,
                new_name=new_name,
            )

            entry = runPy2dsc(filename, new_name)

            break
    else:
        assert False

    os.chdir("deb_dist")
    os.chdir(entry)

    # Build the debian package, but disable the running of tests, will be done later
    # in the pbuilder test steps.
    assert os.system("debuild --set-envvar=DEB_BUILD_OPTIONS=nocheck") == 0

    os.chdir("../../..")
    assert os.path.exists("dist/deb_dist")

    # Cleanup the build directory, not needed anymore.
    shutil.rmtree("build", ignore_errors=True)

    my_print("Uploading...", style="blue")
    os.chdir("dist/deb_dist")

    assert os.system("dput mentors *.changes") == 0

    my_print("Finished.", style="blue")



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""OpenSUSE Build Service (OSC) upload release tool.

Uploads Nuitka branches adapting the RPM configuration to the different
projects on OSC.
"""

import os
import shutil
import sys

from nuitka.tools.environments.Virtualenv import withVirtualenv
from nuitka.tools.release.Documentation import checkReleaseDocumentation
from nuitka.tools.release.Release import (
    checkBranchName,
    cleanupSourceDistributionState,
)
from nuitka.Tracing import tools_logger
from nuitka.utils.InstalledPythons import findInstalledPython
from nuitka.Version import getNuitkaVersion


def main():
    nuitka_version = getNuitkaVersion()
    branch_name = checkBranchName()

    shutil.rmtree("dist", ignore_errors=True)
    shutil.rmtree("build", ignore_errors=True)

    cleanupSourceDistributionState()

    checkReleaseDocumentation()
    # spell-checker: ignore gztar
    assert os.system("%s setup.py sdist --formats=gztar" % sys.executable) == 0

    # Upload stable releases to OpenSUSE Build Service:
    if (
        branch_name.startswith("release")
        or branch_name.startswith("hotfix")
        or branch_name == "main"
    ):
        osc_project_name = "Nuitka"
        spec_suffix = ""
    elif branch_name == "develop":
        osc_project_name = "Nuitka-Unstable"
        spec_suffix = "-unstable"
    elif branch_name == "factory":
        osc_project_name = "Nuitka-experimental"
        spec_suffix = "-experimental"
    else:
        sys.exit("Skipping OSC for branch '%s'" % branch_name)

    installed_python = findInstalledPython(
        python_versions=("3.10",),
        module_specs=None,
    )

    with withVirtualenv(
        "venv_nuitka",
        logger=tools_logger,
        style="blue",
        python=installed_python.getPythonExe(),
    ) as venv:
        venv.runCommand("python -m pip install osc 'setuptools<68'")

        # Stage the "osc" checkout from the ground up,
        # spell-checker: ignore kayhayen,rpmlintrc,addremove
        venv.runCommand(
            """\
rm -rf osc && mkdir osc && cd osc && \
osc checkout home:kayhayen %(osc_project_name)s && \
rm home:kayhayen/%(osc_project_name)s/* && \
cp ../dist/Nuitka-*.tar.gz home:kayhayen/%(osc_project_name)s/ && \
sed -e s/PROJECT_VERSION/%(nuitka_version)s/ ../nuitka/tools/release/rpm/nuitka.spec \
    >home:kayhayen/%(osc_project_name)s/nuitka%(spec_suffix)s.spec && \
sed -i home:kayhayen/%(osc_project_name)s/nuitka%(spec_suffix)s.spec -e \
    's/Name: *PROJECT_NAME/Name: nuitka%(spec_suffix)s/' && \
cp ../nuitka/tools/release/rpm/nuitka-rpmlintrc home:kayhayen/%(osc_project_name)s/ && \
cd home:kayhayen/%(osc_project_name)s/ && \
osc addremove -r && \
echo 'New release' >ci_message && \
osc ci --file ci_message
"""
            % {
                "nuitka_version": nuitka_version,
                "osc_project_name": osc_project_name,
                "spec_suffix": spec_suffix,
            },
            keep_cwd=True,
        )


if __name__ == "__main__":
    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Generated API documentation for Nuitka source."""

import os
import shutil
import stat
import sys

from nuitka.options.CommandLineOptionsTools import makeOptionsParser
from nuitka.tools.Basics import goHome
from nuitka.Tracing import my_print
from nuitka.utils.Execution import (
    check_call,
    getExecutablePath,
    withEnvironmentPathAdded,
)
from nuitka.utils.FileOperations import getFileContents, withTemporaryFile
from nuitka.utils.Utils import isWin32OrPosixWindows


def main():
    goHome()

    parser = makeOptionsParser(usage=None, epilog=None)

    parser.add_option(
        "--upload",
        action="store_true",
        dest="upload",
        default=False,
        help="""\
Upload to https://nuitka.net/apidoc requires access rights and is done by the
official servers automatically only. Without this, create the local html folder
only.

Default is %default.""",
    )

    options, _positional_args = parser.parse_args()

    shutil.rmtree("html", ignore_errors=True)

    doxygen_path = getExecutablePath("doxygen")

    # Extra ball on Windows, check default installation PATH too.
    if not doxygen_path and isWin32OrPosixWindows():
        with withEnvironmentPathAdded("PATH", r"C:\Program Files\Doxygen\bin"):
            doxygen_path = getExecutablePath("doxygen")

    if not doxygen_path:
        sys.exit("Error, need to install Doxygen and add it to PATH for this to work.")

    try:
        import doxypypy  # pylint: disable=I0021,unused-import,unused-variable
    except ImportError:
        sys.exit("Error, needs to install doxypypy into this Python.")

    with withTemporaryFile(suffix=".doxyfile", delete=False) as doxy_file:
        doxy_config = getFileContents("doc/Doxyfile.template")

        with withTemporaryFile(
            suffix=".bat" if isWin32OrPosixWindows() else ".sh", delete=False
        ) as doxy_batch_file:
            if isWin32OrPosixWindows():
                doxy_batch_file.write(
                    "%s -m doxypypy.doxypypy -a -c %%1" % sys.executable
                )
            else:
                doxy_batch_file.write(
                    "#!/bin/sh\nexec '%s' -m doxypypy.doxypypy -a -c $1"
                    % sys.executable
                )

        doxy_batch_filename = doxy_batch_file.name

        st = os.stat(doxy_batch_filename)
        os.chmod(doxy_batch_filename, st.st_mode | stat.S_IEXEC)

        doxy_config = doxy_config.replace("%DOXYPYPY%", doxy_batch_filename)
        doxy_file.write(doxy_config)

        doxy_filename = doxy_file.name

    my_print("Running doxygen:", style="blue")
    try:
        check_call([doxygen_path, doxy_filename])
    finally:
        os.unlink(doxy_filename)
        os.unlink(doxy_batch_filename)

    # Update the repository on the web site.
    if options.upload:
        assert (
            os.system(
                "rsync -avz --delete html/ --chown www-data root@ssh.nuitka.net:/var/www/apidoc/"
            )
            == 0
        )

    my_print("Finished.", style="blue")


if __name__ == "__main__":
    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Interface to WinDepends on Windows."""

import os

from nuitka.containers.OrderedSets import OrderedSet
from nuitka.options.Options import assumeYesForDownloads
from nuitka.Tracing import inclusion_logger
from nuitka.utils.Download import getCachedDownload
from nuitka.utils.Execution import (
    executeProcess,
    getExecutablePath,
    withEnvironmentPathAdded,
)
from nuitka.utils.FileOperations import (
    areSamePaths,
    deleteFile,
    getExternalUsePath,
    getNormalizedPath,
    getNormalizedPathJoin,
    isFilesystemEncodable,
    withFileLock,
)
from nuitka.utils.Json import loadJsonFromFilename
from nuitka.utils.Utils import getArchitecture

from .DllDependenciesCommon import (
    addMissingMsvcRedistDLL,
    attemptToFindNotFoundDLL,
    isMsvcRedistDllName,
    reportMissingMsvcRedistDLLs,
    shallIgnoreMissingDLL,
)

_win_depends_release = "2512-beta"
_win_depends_dotnet_runtime_name = "Microsoft.WindowsDesktop.App"
_win_depends_dotnet_runtime_major = "10"

_installed_dotnet_core_app_versions = None


def _isWinDependsSupportedArchitecture():
    if getArchitecture() == "x86_64":
        return True

    # This is what Windows actually used, spell-checker: ignore PROCESSOR_ARCHITEW6432
    return (
        os.getenv("PROCESSOR_ARCHITECTURE") == "AMD64"
        or os.getenv("PROCESSOR_ARCHITEW6432") == "AMD64"
    )


def _getInstalledDotNetCoreAppRuntimeVersions():
    global _installed_dotnet_core_app_versions  # singleton, pylint: disable=global-statement

    if _installed_dotnet_core_app_versions is not None:
        return _installed_dotnet_core_app_versions

    result = OrderedSet()

    dotnet_exe = getExecutablePath("dotnet")

    if dotnet_exe is not None:
        process_result = executeProcess(command=(dotnet_exe, "--list-runtimes"))

        if process_result.exit_code == 0:
            for line in _decodeWinDependsProcessOutput(process_result).splitlines():
                line = line.strip()

                if not line.startswith(_win_depends_dotnet_runtime_name + " "):
                    continue

                version = line[len(_win_depends_dotnet_runtime_name) + 1 :].split(
                    " ", 1
                )[0]

                if version:
                    result.add(version)

    program_files = os.getenv("ProgramFiles")

    if program_files is not None:
        dotnet_runtime_dir = getNormalizedPathJoin(
            program_files, "dotnet", "shared", _win_depends_dotnet_runtime_name
        )

        if os.path.isdir(dotnet_runtime_dir):
            for filename in os.listdir(dotnet_runtime_dir):
                fullpath = getNormalizedPathJoin(dotnet_runtime_dir, filename)

                if os.path.isdir(fullpath):
                    result.add(filename)

    _installed_dotnet_core_app_versions = tuple(result)

    return _installed_dotnet_core_app_versions


def _checkWinDependsDotNetRuntime():
    installed_versions = _getInstalledDotNetCoreAppRuntimeVersions()
    expected_prefix = _win_depends_dotnet_runtime_major + "."

    for installed_version in installed_versions:
        if installed_version.startswith(expected_prefix):
            return

    # spell-checker: ignore windepends,winget
    return inclusion_logger.sysexit(
        """\
Error, WinDepends release '%(release)s' requires %(runtime_name)s %(runtime_major)s.x \
on the build machine. Install the .NET %(runtime_major)s runtime, e.g. with \
'winget install -e --id Microsoft.DotNet.DesktopRuntime.%(runtime_major)s'."""
        % {
            "release": _win_depends_release,
            "runtime_name": _win_depends_dotnet_runtime_name,
            "runtime_major": _win_depends_dotnet_runtime_major,
        }
    )


def getWinDependsExePath():
    """Return the path of WinDepends.exe (for Windows).

    Will prompt the user to download if not already cached in AppData
    directory for Nuitka.
    """

    if not _isWinDependsSupportedArchitecture():
        return inclusion_logger.sysexit(
            "Error, WinDepends backend is currently only supported on x64 Windows."
        )

    _checkWinDependsDotNetRuntime()

    return getCachedDownload(
        name="WinDepends",
        url=(
            "https://github.com/hfiref0x/WinDepends/releases/download/"
            "%s/WinDepends_v1.0.0.2512_beta_snapshot.zip" % _win_depends_release
        ),
        is_arch_specific="x64",
        binary="WinDepends.exe",
        unzip=True,
        flatten=True,
        specificity=_win_depends_release,
        message="""\
Nuitka will make use of WinDepends (https://github.com/hfiref0x/WinDepends) tool
to analyze the dependencies of Python extension modules when using the
'--experimental=force-dependencies-windepends' backend.""",
        reject="Nuitka cannot use '--experimental=force-dependencies-windepends' without WinDepends.",
        assume_yes_for_downloads=assumeYesForDownloads(),
        download_ok=True,
    )


def _decodeWinDependsProcessOutput(process_result):
    stdout = process_result.stdout
    stderr = process_result.stderr

    if type(stdout) is bytes:
        stdout = stdout.decode("utf8", "replace")

    if type(stderr) is bytes:
        stderr = stderr.decode("utf8", "replace")

    output = "\n".join(
        part.strip() for part in (stdout, stderr) if part and part.strip()
    )

    return output


def _parseWinDependsOutputData(output_filename, binary_filename):
    data = loadJsonFromFilename(output_filename)

    if type(data) is not dict:
        return inclusion_logger.sysexit(
            "Error, WinDepends failed to produce valid JSON output for binary '%s'."
            % binary_filename
        )

    modules = data.get("Modules")

    if type(modules) is not list:
        return inclusion_logger.sysexit(
            "Error, WinDepends failed to produce expected JSON output for binary '%s'."
            % binary_filename
        )

    return modules


def _handleMissingWinDependsModule(dll_filename):
    dll_name = os.path.basename(dll_filename).lower()

    if isMsvcRedistDllName(dll_name):
        addMissingMsvcRedistDLL(dll_name)

    if dll_name.startswith("python") and dll_name.endswith(".dll"):
        return attemptToFindNotFoundDLL(dll_name)

    return None


def _handleWinDependsModuleData(module_data, binary_filename, result):
    dll_filename = module_data.get("FileName")

    if not dll_filename:
        return

    if module_data.get("FileNotFound", False):
        dll_filename = _handleMissingWinDependsModule(dll_filename)

        if dll_filename is None:
            return

    dll_filename = os.path.normcase(os.path.abspath(dll_filename))
    dll_name = os.path.basename(dll_filename).lower()

    if areSamePaths(dll_filename, binary_filename):
        return

    if shallIgnoreMissingDLL(dll_name):
        return

    assert os.path.isfile(dll_filename), dll_filename

    result.add(getNormalizedPath(dll_filename))


def _parseWinDependsOutput(output_filename, binary_filename):
    modules = _parseWinDependsOutputData(
        output_filename=output_filename, binary_filename=binary_filename
    )

    result = OrderedSet()

    for module_data in modules:
        if type(module_data) is dict:
            _handleWinDependsModuleData(
                module_data=module_data,
                binary_filename=binary_filename,
                result=result,
            )

    return result


def detectDLLsWithWinDepends(binary_filename, source_dir, scan_dirs):
    source_dir = getExternalUsePath(source_dir)
    temp_base_name = os.path.basename(binary_filename)

    if not isFilesystemEncodable(temp_base_name):
        temp_base_name = "win_depends"

    output_filename = getNormalizedPathJoin(
        source_dir, temp_base_name + ".windepends.json"
    )

    deleteFile(output_filename, must_exist=False)

    with withFileLock("Finding out WinDepends path for %s" % binary_filename):
        windepends_exe = getWinDependsExePath()

    scan_path = list(getExternalUsePath(dirname) for dirname in scan_dirs)

    with withEnvironmentPathAdded("PATH", *scan_path, prefix=True):
        process_result = executeProcess(
            command=(
                windepends_exe,
                getExternalUsePath(binary_filename),
                "-o",
                output_filename,
                "-f",
                "json",
                "-q",
                "--no-exports",
                "--no-imports",
            ),
            external_cwd=True,
        )

    if not os.path.exists(output_filename):
        output = _decodeWinDependsProcessOutput(process_result)

        if "You must install or update .NET to run this application." in output:
            return inclusion_logger.sysexit("""\
Error, WinDepends requires the .NET 10 runtime on the build machine, \
but the downloaded WinDepends release could not be started.

%s""" % output)

        if output:
            output = "\n\n" + output

        return inclusion_logger.sysexit(
            "Error, 'WinDepends.exe' failed to produce expected output for binary '%s'.%s"
            % (binary_filename, output)
        )

    result = _parseWinDependsOutput(
        output_filename=output_filename, binary_filename=binary_filename
    )

    deleteFile(output_filename, must_exist=True)

    reportMissingMsvcRedistDLLs()

    return result


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Launcher for git hook installer tool."""

import os
import sys

# Unchanged, running from checkout, use the parent directory, the nuitka
# package ought be there.
sys.path.insert(0, os.path.normpath(os.path.join(os.path.dirname(__file__), "..")))

# isort:start

import stat

from nuitka.options.CommandLineOptionsTools import makeOptionsParser
from nuitka.tools.Basics import goHome
from nuitka.tools.release.Release import getGitDir
from nuitka.Tracing import tools_logger
from nuitka.utils.Execution import getExecutablePath
from nuitka.utils.FileOperations import (
    getFileContents,
    getWindowsShortPathName,
)


def _getWindowsShPath():
    git_path = getExecutablePath("git")

    if git_path is None:
        git_path = r"C:\Program Files\Git\bin\sh.exe"

        if not os.path.exists(git_path):
            git_path = None

    if git_path is None:
        return tools_logger.sysexit("""\
Error, cannot locate 'git.exe' which we need to install git hooks. Add it to
PATH while executing this will be sufficient.""")

    for candidate in (
        "sh.exe",
        os.path.join("..", "bin", "sh.exe"),
        os.path.join("..", "..", "bin", "sh.exe"),
    ):
        result = os.path.normpath(os.path.join(os.path.dirname(git_path), candidate))

        if os.path.exists(result):
            break
    else:
        return tools_logger.sysexit("""\
Error, cannot locate 'sh.exe' near 'git.exe' which we need to install git hooks,
please improve this script.""")

    # For MinGW and #! we will need a path without spaces, so use this
    # code to find the short name, that won't have it.
    return getWindowsShortPathName(result)


def main():
    parser = makeOptionsParser(usage="usage: %prog [options]", epilog="")

    parser.add_option(
        "--no-pre-commit",
        action="store_false",
        dest="pre_commit",
        default=True,
        help="Do not install the pre-commit hook.",
    )
    parser.add_option(
        "--no-pre-push",
        action="store_false",
        dest="pre_push",
        default=True,
        help="Do not install the pre-push hook.",
    )
    options, _args = parser.parse_args()

    goHome()

    git_dir = getGitDir()

    for hook in os.listdir(".githooks"):
        hook_target = os.path.join(git_dir, "hooks", hook)

        if hook == "pre-commit" and not options.pre_commit:
            if os.path.exists(hook_target):
                os.remove(hook_target)
            continue
        if hook == "pre-push" and not options.pre_push:
            if os.path.exists(hook_target):
                os.remove(hook_target)
            continue

        full_path = os.path.join(".githooks", hook)

        hook_contents = getFileContents(full_path)

        if hook_contents.startswith("#!/bin/sh"):
            if os.name == "nt":
                sh_path = _getWindowsShPath()
                # Correct shebang for Windows git to work.
                hook_contents = "#!%s\n%s" % (
                    sh_path.replace("\\", "/").replace(" ", r"\ "),
                    hook_contents[10:],
                )

            # Also use "sys.executable" to make sure we find Nuitka tools.
            for tool_name in (
                "autoformat-nuitka-source",
                "check-nuitka-with-pylint",
                "generate-specialized-c-code",
                "generate-specialized-python-code",
            ):
                hook_contents = hook_contents.replace(
                    "./bin/%s" % tool_name,
                    "'%s' ./bin/%s" % (sys.executable, tool_name),
                )
        else:
            sys.exit("Error, unknown hook contents.")

        with open(hook_target, "wb") as out_file:
            out_file.write(hook_contents.encode("utf8"))

        st = os.stat(hook_target)
        os.chmod(hook_target, st.st_mode | stat.S_IEXEC)


if __name__ == "__main__":
    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Test Nuitka compiling itself and compiling itself in compiled form again.

This should not only give no errors, but the same source for modules being
compiled when Nuitka is running compiled and uncompiled, so we can discover
changes in order of execution in this test.
"""

import os
import sys

# Find nuitka package relative to us.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
    ),
)

# isort:start

import shutil
import subprocess

from nuitka.build.SconsInterface import getInlineSconsLibPath
from nuitka.options.CommandLineOptionsTools import makeOptionsParser
from nuitka.reports.CompilationReportReader import (
    getCompilationOutputBinary,
    parseCompilationReport,
)
from nuitka.tools.Basics import addPYTHONPATH
from nuitka.tools.testing.Common import (
    getPythonSysPath,
    getTempDir,
    my_print,
    setup,
    test_logger,
    withPythonPathChange,
)
from nuitka.utils.Diffs import compareDirectories
from nuitka.utils.Execution import wrapCommandForDebuggerForSubprocess
from nuitka.utils.FileOperations import (
    copyTree,
    deleteFile,
    getFileList,
    getSubDirectories,
    listDir,
    removeDirectory,
)
from nuitka.utils.Importing import getExtensionModuleSuffix
from nuitka.Version import getCommercialVersion

nuitka_main_path = os.path.join("..", "..", "bin", "nuitka")

tmp_dir = getTempDir()

# Cannot detect this more automatic, so we need to list them, avoiding
# the ones not needed.
PACKAGE_LIST = [
    "nuitka",
    "nuitka/build",
    "nuitka/code_generation",
    "nuitka/code_generation/c_types",
    "nuitka/code_generation/templates",
    "nuitka/containers",
    "nuitka/finalizations",
    "nuitka/freezer",
    "nuitka/importing",
    "nuitka/nodes",
    "nuitka/nodes/shapes",
    "nuitka/optimizations",
    "nuitka/options",
    "nuitka/pgo",
    "nuitka/plugins",
    "nuitka/plugins/commercial",
    "nuitka/plugins/standard",
    "nuitka/reports",
    "nuitka/specs",
    "nuitka/tree",
    "nuitka/format",
    "nuitka/package_config",
    "nuitka/utils",
]

if not getCommercialVersion():
    PACKAGE_LIST.remove("nuitka/plugins/commercial")

exe_suffix = ".exe" if os.name == "nt" else ".bin"
nuitka_runner_binary = "nuitka-runner" + exe_suffix
data_composer_main = "DataComposer.py"
scons_main = "scons.py"
pass_numbers = (1, 2, 3, 4, 5)


def _traceCompilation(path, pass_number):
    test_logger.info("Compiling '%s' (PASS %d)." % (path, pass_number))


def _updateDataComposerMain(filename, use_checkout_path):
    with open(filename, "w") as output_file:
        if use_checkout_path:
            output_file.write("""\
#!/usr/bin/env python

import os
import sys

sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
    ),
)

from nuitka.tools.data_composer.DataComposer import main

del sys.path[0]

sys.path = [
    path_element
    for path_element in sys.path
    if os.path.dirname(os.path.abspath(__file__)) != path_element
]

main()
""")
        else:
            output_file.write("""\
#!/usr/bin/env python

from nuitka.tools.data_composer.DataComposer import main

main()
""")


def _updateNuitkaRunnerMain():
    with open("nuitka-runner.py", "w") as output_file:
        output_file.write("""\
#!/usr/bin/env python

\"\"\"Launcher for reflected Nuitka compiler runs.\"\"\"

import os
import sys

sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
    ),
)

import nuitka.__main__  # false alarm, pylint: disable=I0021,no-name-in-module

del sys.path[0]

sys.path = [
    path_element
    for path_element in sys.path
    if os.path.dirname(os.path.abspath(__file__)) != path_element
]

nuitka.__main__.main()
""")


def _updateSconsMain(filename):
    with open(filename, "w") as output_file:
        output_file.write("""\
#!/usr/bin/env python

import os

import SCons
import SCons.Script

os.environ["SCONS_LIB_DIR"] = os.path.dirname(SCons.__file__)

SCons.Script.main()
""")


def _createMultidistBinaryAlias(binary_path, alias_name):
    alias_path = os.path.join(os.path.dirname(binary_path), alias_name + exe_suffix)

    deleteFile(alias_path, must_exist=False)

    try:
        os.link(binary_path, alias_path)
    except OSError:
        shutil.copy2(binary_path, alias_path)

    return alias_path


def _parsePassNumber(pass_number, option_name):
    try:
        pass_number = int(pass_number)
    except ValueError:
        return test_logger.sysexit(
            "The '%s' option got invalid pass number '%s'." % (option_name, pass_number)
        )

    if pass_number not in pass_numbers:
        return test_logger.sysexit(
            "The '%s' option got pass number '%s' outside supported range %d-%d."
            % (option_name, pass_number, pass_numbers[0], pass_numbers[-1])
        )

    return pass_number


def _parsePassSpec(pass_spec, option_name):
    result = set()

    for pass_item in pass_spec.split(","):
        pass_item = pass_item.strip()

        if not pass_item:
            continue

        if "-" in pass_item:
            start_pass, stop_pass = pass_item.split("-", 1)

            start_pass = _parsePassNumber(start_pass, option_name)
            stop_pass = _parsePassNumber(stop_pass, option_name)

            if start_pass > stop_pass:
                return test_logger.sysexit(
                    "The '%s' option got descending pass range '%s'."
                    % (option_name, pass_item)
                )

            for pass_number in range(start_pass, stop_pass + 1):
                result.add(pass_number)
        else:
            result.add(_parsePassNumber(pass_item, option_name))

    if not result:
        return test_logger.sysexit(
            "The '%s' option needs at least one pass number." % option_name
        )

    return result


def _recordPassSelection(option, opt_str, value, parser, operation):
    pass_selection_operations = parser.values.pass_selection_operations

    if pass_selection_operations is None:
        pass_selection_operations = []
        parser.values.pass_selection_operations = pass_selection_operations

    pass_selection_operations.append((operation, _parsePassSpec(value, opt_str)))


def _parseArgs():
    parser = makeOptionsParser(usage=None, epilog=None)
    parser.error = test_logger.sysexit

    parser.set_defaults(pass_selection_operations=None)

    parser.add_option(
        "--run-passes",
        action="callback",
        type="string",
        callback=_recordPassSelection,
        callback_args=("run",),
        help="""\
Run only the selected passes, e.g. '1,3-5'. Can be given multiple times.""",
    )

    parser.add_option(
        "--skip-passes",
        action="callback",
        type="string",
        callback=_recordPassSelection,
        callback_args=("skip",),
        help="""\
Skip selected passes from the current pass set, e.g. '2,4'. Can be given multiple times.""",
    )

    options, positional_args = parser.parse_args()

    enabled_passes = set(pass_numbers)

    if options.pass_selection_operations:
        for operation, selected_passes in options.pass_selection_operations:
            if operation == "run":
                enabled_passes = selected_passes
            else:
                enabled_passes.difference_update(selected_passes)

    return enabled_passes, positional_args


def executePASS1():
    test_logger.info(
        "PASS 1: Compiling to many compiled modules from compiler running from .py files."
    )

    base_dir = os.path.join("..", "..")

    for package in PACKAGE_LIST:
        package = package.replace("/", os.path.sep)

        source_dir = os.path.join(base_dir, package)
        target_dir = package

        removeDirectory(
            path=target_dir,
            logger=test_logger,
            ignore_errors=False,
            extra_recommendation=None,
        )

        os.mkdir(target_dir)

        for path, filename in listDir(target_dir):
            if filename.endswith((".so", ".dylib")):
                os.unlink(path)

        for path, filename in listDir(source_dir):
            if not filename.endswith(".py"):
                continue

            if filename.startswith(".#"):
                continue

            if filename != "__init__.py":
                _traceCompilation(path=path, pass_number=1)

                command = [
                    os.environ["PYTHON"],
                    nuitka_main_path,
                    "--mode=module",
                    "--nofollow-imports",
                    "--output-dir=%s" % target_dir,
                    "--no-pyi-file",
                    path,
                ]
                command += os.getenv("NUITKA_EXTRA_OPTIONS", "").split()

                my_print("Command: ", " ".join(command))

                result = subprocess.call(command)

                if result != 0:
                    sys.exit(result)
            else:
                shutil.copyfile(path, os.path.join(target_dir, filename))

    _traceCompilation(path=nuitka_main_path, pass_number=1)

    _updateNuitkaRunnerMain()
    _updateDataComposerMain(data_composer_main, use_checkout_path=True)

    command = [
        os.environ["PYTHON"],
        nuitka_main_path,
        "--nofollow-imports",
        "--output-dir=.",
        "--python-flag=no_site",
        "--main=nuitka-runner.py",
        "--main=%s" % data_composer_main,
    ]
    command += os.getenv("NUITKA_EXTRA_OPTIONS", "").split()

    my_print("Command: ", " ".join(command))

    try:
        result = subprocess.call(command)
    finally:
        deleteFile(path=data_composer_main, must_exist=False)

    if result != 0:
        sys.exit(result)

    # Keep the binary name aligned with the multidist entry point name.
    deleteFile(path="nuitka" + exe_suffix, must_exist=False)

    scons_inline_copy_path = os.path.join(base_dir, "nuitka", "build", "inline_copy")

    if os.path.exists(scons_inline_copy_path):
        copyTree(scons_inline_copy_path, os.path.join("nuitka", "build", "inline_copy"))

    # Copy required data files.
    for filename in (
        "nuitka/build/Backend.scons",
        "nuitka/build/CCompilerVersion.scons",
        "nuitka/build/Offsets.scons",
        "nuitka/build/Onefile.scons",
        "nuitka/plugins/standard/standard.nuitka-package.config.yml",
        "nuitka/plugins/standard/stdlib3.nuitka-package.config.yml",
        "nuitka/plugins/standard/stdlib2.nuitka-package.config.yml",
    ):
        shutil.copyfile(
            os.path.join(base_dir, filename),
            filename,
        )

    copyTree(
        os.path.join(base_dir, "nuitka", "code_generation", "templates_c"),
        os.path.join("nuitka", "code_generation", "templates_c"),
    )

    copyTree(
        os.path.join(base_dir, "nuitka", "build", "static_src"),
        os.path.join("nuitka", "build", "static_src"),
    )
    copyTree(
        os.path.join(base_dir, "nuitka", "build", "include"),
        os.path.join("nuitka", "build", "include"),
    )

    # The data composer tool, use it by source.
    copyTree(
        os.path.join(base_dir, "nuitka", "tools"),
        os.path.join("nuitka", "tools"),
    )

    test_logger.info("OK.")


def compileAndCompareWith(nuitka, pass_number):
    if "PYTHONHASHSEED" not in os.environ:
        os.environ["PYTHONHASHSEED"] = "0"
    if "PYTHON_FROZEN_MODULES" not in os.environ:
        os.environ["PYTHON_FROZEN_MODULES"] = "off"

    base_dir = os.path.join("..", "..")

    for package in PACKAGE_LIST:
        package = package.replace("/", os.path.sep)

        source_dir = os.path.join(base_dir, package)

        for path, filename in listDir(source_dir):
            if not filename.endswith(".py"):
                continue

            if filename.startswith(".#"):
                continue

            path = os.path.join(source_dir, filename)

            if filename != "__init__.py":
                _traceCompilation(path=path, pass_number=pass_number)

                target = filename.replace(".py", ".build")

                target_dir = os.path.join(tmp_dir, target)

                removeDirectory(
                    logger=test_logger,
                    path=target_dir,
                    ignore_errors=False,
                    extra_recommendation=None,
                )

                command = [
                    nuitka,
                    "--mode=module",
                    "--output-dir=%s" % tmp_dir,
                    "--no-pyi-file",
                    "--nofollow-imports",
                    path,
                ]
                command += os.getenv("NUITKA_EXTRA_OPTIONS", "").split()

                my_print("Command: ", " ".join(command))
                exit_nuitka = subprocess.call(command)

                # In case of segfault or assertion triggered, run in debugger.
                if exit_nuitka in (-11, -6) and sys.platform != "nt":
                    command2 = wrapCommandForDebuggerForSubprocess(
                        command=command, debugger=os.getenv("NUITKA_DEBUGGER_CHOICE")
                    )
                    subprocess.call(command2)

                if exit_nuitka != 0:
                    my_print("An error exit %s occurred, aborting." % exit_nuitka)
                    sys.exit(exit_nuitka)

                has_diff = compareDirectories(
                    logger=test_logger,
                    my_print=my_print,
                    dir1=os.path.join(package, target),
                    dir2=target_dir,
                    ignore_suffixes=(
                        ".o",
                        ".os",
                        ".obj",
                        ".dblite",  # spell-checker: ignore dblite
                        ".tmp",
                        ".sconsign",  # spell-checker: ignore sconsign
                        ".txt",
                        ".bin",
                        ".const",
                        ".exp",
                        ".scons",
                    ),
                    ignore_substrings=(
                        "cache-",
                        "scons-debug",
                    ),
                )

                if has_diff:
                    sys.exit("There were differences!")

                shutil.rmtree(target_dir)

                for preferred in (True, False):
                    target_filename = filename.replace(
                        ".py", getExtensionModuleSuffix(preferred=preferred)
                    )

                    deleteFile(
                        path=os.path.join(tmp_dir, target_filename), must_exist=False
                    )


def executePASS2():
    test_logger.info(
        "PASS 2: Compiling from compiler running from entry '.exe' and many extension files."
    )

    with withPythonPathChange(getPythonSysPath()):
        # Windows will load the compiled modules (pyd) only from PYTHONPATH, so we
        # have to add it.
        if os.name == "nt":
            addPYTHONPATH(PACKAGE_LIST)

        compileAndCompareWith(
            nuitka=os.path.join(".", nuitka_runner_binary), pass_number=2
        )

    # Cleanup compiled modules that would otherwise confuse PASS3, but keep the
    # PASS1 ".build" directories as comparison baseline for PASS4 and repeated
    # late-pass reruns.
    for filename in getFileList("nuitka", only_suffixes=(".so", ".pyd")):
        deleteFile(filename, must_exist=True)

    test_logger.info("OK.")


def executePASS3():
    test_logger.info(
        "PASS 3: Compiling from compiler running from .py files to single .exe."
    )

    exe_path = os.path.join(tmp_dir, "nuitka-runner" + exe_suffix)

    if os.path.exists(exe_path):
        os.unlink(exe_path)

    build_path = os.path.join(tmp_dir, "nuitka-runner.build")

    if os.path.exists(build_path):
        shutil.rmtree(build_path)

    base_dir = os.path.abspath(os.path.join("..", ".."))
    helper_dir = os.path.join(tmp_dir, "pass3-main")

    if not os.path.exists(helper_dir):
        os.makedirs(helper_dir)

    runner_main_path = os.path.join(helper_dir, "nuitka-runner.py")
    data_composer_main_path = os.path.join(helper_dir, data_composer_main)
    scons_main_path = os.path.join(helper_dir, scons_main)

    path = os.path.join("..", "..", "bin", "nuitka")

    _traceCompilation(path=path, pass_number=3)

    with withPythonPathChange((base_dir, getInlineSconsLibPath())):
        with open(runner_main_path, "w") as output_file:
            output_file.write("""\
#!/usr/bin/env python

\"\"\"Launcher for source-based Nuitka compiler runs.\"\"\"

import nuitka.__main__  # false alarm, pylint: disable=I0021,no-name-in-module

nuitka.__main__.main()
""")

        _updateDataComposerMain(data_composer_main_path, use_checkout_path=False)
        _updateSconsMain(scons_main_path)

        command = [
            os.environ["PYTHON"],
            os.path.join(base_dir, "bin", "nuitka"),
            "--output-dir=%s" % tmp_dir,
            "--python-flag=-S",
            "--follow-imports",
            "--include-package=nuitka.plugins.standard",
            "--nofollow-import-to=*-postLoad",
            "--nofollow-import-to=pip",
            "--report=%s" % os.path.abspath("compilation-report-pass3.xml"),
            "--main=%s" % runner_main_path,
            "--main=%s" % data_composer_main_path,
            "--main=%s" % scons_main_path,
        ]

        my_print("Command: ", " ".join(command))
        result = subprocess.call(command, cwd=base_dir)

    if result != 0:
        sys.exit(result)

    _createMultidistBinaryAlias(exe_path, "scons")

    shutil.rmtree(build_path)

    test_logger.info("OK.")


def executePASS4():
    test_logger.info("PASS 4: Compiling the compiler running from single exe.")

    compilation_report = parseCompilationReport("compilation-report-pass3.xml")

    exe_path = getCompilationOutputBinary(
        compilation_report=compilation_report,
        prefixes=(("${cwd}", os.getcwd()),),
    )

    # The source compiler uses the host Python import path for optional
    # non-stdlib modules like "pkg_resources", so mirror that here to keep
    # import lowering decisions stable for comparisons.
    with withPythonPathChange(
        getPythonSysPath().split(os.pathsep) + [os.path.join("..", "..")]
    ):
        compileAndCompareWith(exe_path, pass_number=4)

    test_logger.info("OK.")


def executePASS5():
    my_print(
        "PASS 5: Compiling the compiler 'nuitka' package to single extension module."
    )

    path = os.path.join("..", "..", "nuitka")

    command = [
        os.environ["PYTHON"],
        nuitka_main_path,
        "--output-dir=%s" % tmp_dir,
        "--include-plugin-dir=%s" % path,
        "--nofollow-import-to=nuitka.build.inline_copy",
        "--nofollow-import-to=nuitka.build.include",
        "--nofollow-import-to=nuitka.build.static_src",
        "--nofollow-import-to=nuitka.tools",
        "--mode=module",
        path,
    ]

    result = subprocess.call(command)

    if result != 0:
        sys.exit(result)

    for preferred in (True, False):
        candidate = "nuitka" + getExtensionModuleSuffix(preferred=preferred)

        deleteFile(candidate, must_exist=False)

    os.unlink(os.path.join(tmp_dir, "nuitka.pyi"))
    shutil.rmtree(os.path.join(tmp_dir, "nuitka.build"))


def main():
    enabled_passes, _unused_positional_args = _parseArgs()

    setup(needs_io_encoding=True)

    if enabled_passes != set(pass_numbers):
        selected_passes = (
            ",".join(str(pass_number) for pass_number in sorted(enabled_passes))
            or "none"
        )
        test_logger.info("Selected passes: %s." % selected_passes)

    for pass_number, pass_function in (
        (1, executePASS1),
        (2, executePASS2),
        (3, executePASS3),
        (4, executePASS4),
        (5, executePASS5),
    ):
        if pass_number not in enabled_passes:
            test_logger.info("PASS %d: Skipped by option." % pass_number)
            continue

        pass_function()


if __name__ == "__main__":
    main()

#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


from __future__ import print_function

import itertools


class C:
    def compiled_method(self, a=1, b=2, c=3, d=4, e=5, f=6):
        return a, b, c, d, e, f


def calledRepeatedly():
    inst = C()

    # This is supposed to make a call to a compiled method, which is
    # being optimized separately.
    # construct_begin
    inst.compiled_method()
    inst.compiled_method()
    inst.compiled_method()
    # construct_alternative
    pass
    # construct_end


for x in itertools.repeat(None, 50000):
    calledRepeatedly()

print("OK.")

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Handling of fake module descriptions and reserved names."""

from nuitka.containers.Namedtuples import makeNamedtupleClass
from nuitka.containers.OrderedSets import OrderedSet
from nuitka.utils.ModuleNames import ModuleName

from .ImportingResults import makeFindModuleResult

FakeModuleDescription = makeNamedtupleClass(
    "FakeModuleDescription",
    (
        "module_name",
        "source_code",
        "source_filename",
        "reason",
    ),
)

_fake_modules = OrderedSet()


def makeFakeModuleDescription(module_name, source_code, source_filename, reason):
    return FakeModuleDescription(
        module_name=module_name,
        source_code=source_code,
        source_filename=source_filename,
        reason=reason,
    )


def addFakeModule(module_name):
    assert type(module_name) is ModuleName, module_name

    _fake_modules.add(module_name)


def locateFakeModule(module_name):
    if module_name in _fake_modules:
        return makeFindModuleResult(
            found_module_name=module_name,
            module_package=module_name.getPackageName(),
            module_filename=None,
            module_kind="py",
            finding="fake",
        )

    return None


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Runner for distutils integration

Tests for example packages demonstrating that wheel creation with Nuitka
is compatible to standard packaging tools.

"""

import os
import sys

# Find nuitka package relative to us. The replacement is for POSIX python
# and Windows paths on command line.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__.replace("\\", os.sep))), "..", ".."
        )
    ),
)

# isort:start

import subprocess

from nuitka.reports.CompilationReportReader import (
    getCompilationOutputBinary,
    parseCompilationReport,
)
from nuitka.tools.environments.Virtualenv import (
    NuitkaCalledProcessError,
    withVirtualenv,
)
from nuitka.tools.testing.Common import (
    createSearchMode,
    getPythonBinary,
    my_print,
    reportSkip,
    scanDirectoryForTestCaseFolders,
    setup,
    test_logger,
)
from nuitka.tools.testing.OutputComparison import compareOutput
from nuitka.utils.Execution import wrapCommandForDebuggerForSubprocess
from nuitka.utils.FileOperations import (
    deleteFile,
    getFileContents,
    putTextFileContents,
    removeDirectory,
)


def _adaptPyProjectFile(case_dir, variant):
    assert variant in ("cpython", "nuitka"), variant

    source_filename = "pyproject.%s.toml" % variant

    pyproject_filename = os.path.join(case_dir, "pyproject.toml")

    contents = getFileContents(os.path.join(case_dir, source_filename)).replace(
        "file:../../..",
        "file:%s"
        % (
            os.path.abspath(os.path.join(case_dir, "..", "..", "..")).replace("\\", "/")
        ),
    )

    extra_options = []
    nuitka_extra_options = os.getenv("NUITKA_EXTRA_OPTIONS")

    if nuitka_extra_options and variant == "nuitka":
        extra_options.extend(nuitka_extra_options.split())

    if extra_options:
        # May need to add it.
        assert "[tool.nuitka]" in contents

        contents += "\n"

        for extra_option in extra_options:
            assert extra_option.startswith("--"), extra_option

            if "=" not in extra_option:
                extra_option += "=true"

            contents += "%s\n" % extra_option

    putTextFileContents(
        filename=pyproject_filename,
        contents=contents,
    )

    return pyproject_filename


def installBuildTools(venv, filename, is_nuitka):
    install_command = "pip install "

    if is_nuitka:
        install_command += "setuptools "

    install_command += "build "

    if "_uv_" in filename:
        install_command += "uv_build"
    elif "_poetry_" in filename:
        install_command += "poetry-core"
    else:
        install_command += "setuptools wheel"

    venv.runCommand(
        install_command,
        style="test-prepare",
    )


def _handleCase(python_version, nuitka_dir, filename):
    # Complex stuff, pylint: disable=too-many-branches,too-many-locals,too-many-statements

    my_print("Consider distutils example: '%s'" % filename, style="blue")

    if "dependencies" in filename:
        # TODO: Dependencies are not yet supported, but need to create a virtualenv in Nuitka
        # during compilation to achieve that, but it's not done yet.
        reportSkip(
            "Skipped, dependencies are not yet supported",
            ".",
            filename,
        )
        return

    is_pyproject = "_pyproject" in filename

    if is_pyproject and python_version < (3,):
        reportSkip(
            "Skipped, pyproject cases are not for Python2",
            ".",
            filename,
        )
        return

    case_dir = os.path.join(os.getcwd(), filename)

    removeDirectory(
        os.path.join(case_dir, "build"),
        logger=test_logger,
        ignore_errors=False,
        extra_recommendation=None,
    )
    removeDirectory(
        os.path.join(case_dir, "dist"),
        logger=test_logger,
        ignore_errors=False,
        extra_recommendation=None,
    )

    with withVirtualenv(
        "venv_cpython", python=getPythonBinary(), logger=test_logger
    ) as venv:
        installBuildTools(
            venv=venv,
            filename=filename,
            is_nuitka=False,
        )

        if is_pyproject:
            pyproject_filename = _adaptPyProjectFile(
                case_dir=case_dir, variant="cpython"
            )

            venv.runCommand(
                commands=['cd "%s"' % case_dir, "python -m build"], style="test-prepare"
            )
            deleteFile(pyproject_filename, must_exist=True)
        else:
            venv.runCommand(
                commands=['cd "%s"' % case_dir, "python setup.py bdist_wheel"],
                style="test-prepare",
            )

        dist_dir = os.path.join(case_dir, "dist")

        venv.runCommand(
            'pip install "%s"' % (os.path.join(dist_dir, os.listdir(dist_dir)[0]))
        )

        runner_binary = venv.getBinaryPath("runner")
        # pylint: disable=consider-using-with

        if os.path.exists(runner_binary):
            # Need to call CPython binary for Windows.
            process = subprocess.Popen(
                args=[
                    venv.getBinaryPath("python"),
                    venv.getBinaryPath("runner"),
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        else:
            assert os.path.exists(runner_binary + ".exe")

            process = subprocess.Popen(
                args=[runner_binary + ".exe"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

        stdout_cpython, stderr_cpython = process.communicate()
        exit_cpython = process.returncode

        my_print("STDOUT CPython:")
        my_print(stdout_cpython)
        my_print("STDERR CPython:")
        my_print(stderr_cpython)

        assert exit_cpython == 0, exit_cpython
        my_print("EXIT was OK.")

    removeDirectory(
        os.path.join(case_dir, "build"),
        logger=test_logger,
        ignore_errors=False,
        extra_recommendation=None,
    )
    removeDirectory(
        os.path.join(case_dir, "dist"),
        logger=test_logger,
        ignore_errors=False,
        extra_recommendation=None,
    )

    with withVirtualenv(
        "venv_nuitka", python=getPythonBinary(), logger=test_logger
    ) as venv:
        installBuildTools(
            venv=venv,
            filename=filename,
            is_nuitka=True,
        )

        venv.runCommand(
            commands=['cd "%s"' % nuitka_dir, "python setup.py install"],
            style="test-prepare",
            env={"PYTHONWARNINGS": "ignore"},
        )
        # Remove that left over from the install command.
        removeDirectory(
            path=os.path.join(nuitka_dir, "Nuitka.egg-info"),
            logger=test_logger,
            ignore_errors=False,
            extra_recommendation=None,
        )

        build_commands = []

        if not is_pyproject:
            build_commands.append("python setup.py bdist_nuitka")

        if "versioneer" not in case_dir:  # spell-checker: ignore versioneer
            build_commands.append(
                "python -m nuitka --project --mode=standalone --report=compilation-report.xml"
            )

        for command_desc in build_commands:
            # Create the wheel or standalone binary with Nuitka compilation.
            if is_pyproject:
                pyproject_filename = _adaptPyProjectFile(
                    case_dir=case_dir, variant="nuitka"
                )

                venv.runCommand(
                    commands=['cd "%s"' % case_dir, command_desc],
                    style="test-execution",
                )

                deleteFile(pyproject_filename, must_exist=True)
            else:
                venv.runCommand(
                    commands=['cd "%s"' % case_dir, command_desc],
                    style="test-execution",
                )

            if "--project" not in command_desc:
                dist_dir = os.path.join(case_dir, "dist")
                venv.runCommand(
                    'pip install "%s"'
                    % (os.path.join(dist_dir, os.listdir(dist_dir)[0]))
                )

                runner_binary = os.path.join(
                    venv.getVirtualenvDir(),
                    "bin" if os.name != "nt" else "scripts",
                    "runner",
                )

                # TODO: Is this something to be abstracted into a function.
                # pylint: disable=consider-using-with
                if os.path.exists(runner_binary):
                    command = [
                        os.path.join(
                            venv.getVirtualenvDir(),
                            "bin" if os.name != "nt" else "scripts",
                            "python",
                        ),
                        runner_binary,
                    ]
                else:
                    assert os.path.exists(runner_binary + ".exe")

                    command = [runner_binary + ".exe"]

                process = subprocess.Popen(
                    args=command,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )

                stdout_nuitka, stderr_nuitka = process.communicate()
                exit_nuitka = process.returncode

                if exit_nuitka == -11:
                    command = wrapCommandForDebuggerForSubprocess(
                        command=command, debugger=os.getenv("NUITKA_DEBUGGER_CHOICE")
                    )

                    test_logger.warning("Rerunning due to segfault in debugger.")
                    stdout_nuitka, stderr_nuitka = process.communicate()

                    process = subprocess.Popen(
                        args=command,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                    )

                    stdout_nuitka, stderr_nuitka = process.communicate()

                my_print("STDOUT Nuitka:")
                my_print(stdout_nuitka)
                my_print("STDERR Nuitka:")
                my_print(stderr_nuitka)

                if exit_nuitka == -11:
                    my_print("EXIT was segfault.")
                    os._exit(2)

                assert exit_nuitka == 0, exit_nuitka
                my_print("EXIT was OK.")
            else:
                my_print("Matching output of '%s' against CPython..." % command_desc)

                compilation_report = parseCompilationReport(
                    os.path.join(case_dir, "compilation-report.xml")
                )

                binary_filename = getCompilationOutputBinary(
                    compilation_report=compilation_report,
                    prefixes=(("${cwd}", case_dir),),
                )

                process = subprocess.Popen(
                    args=[binary_filename],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )

                stdout_nuitka, stderr_nuitka = process.communicate()
                exit_nuitka = process.returncode

            exit_code_stdout = compareOutput(
                "stdout",
                stdout_cpython,
                stdout_nuitka,
                ignore_warnings=True,
                syntax_errors=True,
            )

            exit_code_stderr = compareOutput(
                "stderr",
                stderr_cpython,
                stderr_nuitka,
                ignore_warnings=True,
                syntax_errors=True,
            )

            exit_code_return = exit_cpython != exit_nuitka

            if exit_code_return:
                my_print(
                    """\
Exit codes {exit_cpython:d} (CPython) != {exit_nuitka:d} (Nuitka)""".format(
                        exit_cpython=exit_cpython, exit_nuitka=exit_nuitka
                    )
                )

            exit_code = exit_code_stdout or exit_code_stderr or exit_code_return

            if exit_code:
                sys.exit("Error, outputs differed.")

            my_print("OK, outputs matched.")


def main():
    python_version = setup(suite="distutils", needs_io_encoding=True)

    search_mode = createSearchMode()

    nuitka_dir = os.path.abspath(os.path.join(os.getcwd(), "..", ".."))

    for filename, _filename_main in scanDirectoryForTestCaseFolders("."):
        active = search_mode.consider(dirname=None, filename=filename)

        if active:
            try:
                _handleCase(python_version, nuitka_dir, filename)
            except NuitkaCalledProcessError:
                test_logger.sysexit("Error in test case '%s'." % filename)

    search_mode.finish()


if __name__ == "__main__":
    main()

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
